# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from ._base import Field
from caterpillar.abc import (
    _ContextLambda,
    _ContextLike,
    _StructLike,
    _SwitchLike,
    _LengthT,
    _IT,
    _OT,
    _OptionLike,
)
from caterpillar.byteorder import ByteOrder
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Generic,
    List,
    Type,
    TypeVar,
    Union,
    overload,
)

_NextOT = TypeVar("_NextOT")

class FieldMixin(Generic[_IT, _OT]):
    def __or__(self, flag: _OptionLike) -> Field[_IT, _OT]: ...
    def __xor__(self, flag: _OptionLike) -> Field[_IT, _OT]: ...
    def __matmul__(self, offset: _ContextLambda | int) -> Field[_IT, _OT]: ...
    def __getitem__(self, dim: _LengthT) -> Field[Collection[_IT], Collection[_OT]]: ...
    def __rshift__(self, switch: _SwitchLike) -> Field[_IT, _OT]: ...
    def __floordiv__(self, condition: _ContextLambda | bool) -> Field[_IT, _OT]: ...
    def __set_byteorder__(self, order: ByteOrder) -> Field[_IT, _OT]: ...
    def __rsub__(self, bits: _ContextLambda | int) -> Field[_IT, _OT]: ...
    def __and__(self, other: _StructLike) -> Chain: ...

class FieldStruct(FieldMixin[_IT, _OT], _StructLike[_IT, _OT]):
    def pack_single(self, obj: _IT, context: _ContextLike) -> None: ...
    def unpack_single(self, context: _ContextLike) -> _OT: ...
    def pack_seq(self, seq: Collection[_IT], context: _ContextLike) -> None: ...
    def unpack_seq(self, context: _ContextLike) -> Collection[_OT]: ...
    def __pack__(self, obj: _IT, context: _ContextLike) -> None: ...
    def __unpack__(self, context: _ContextLike) -> _OT: ...

class Chain(FieldStruct[_IT, _OT]):
    @overload
    def __init__(
        self, initial: _StructLike[_IT, Any], *structs: _StructLike
    ) -> None: ...
    @overload
    def __init__(self, initial: _StructLike[_IT, _OT], *structs: None) -> None: ...
    @property
    def head(self) -> _StructLike[_IT, Any]: ...
    @property
    def tail(self) -> _StructLike[Any, _OT]: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def __type__(self) -> Type[_OT]: ...
    def __and__(self, other: _StructLike[_OT, _NextOT]) -> Chain[_IT, _NextOT]: ...
    def __rand__(self, other: _StructLike[_OT, _NextOT]) -> Chain[_IT, _NextOT]: ...
    def unpack_single(self, context: _ContextLike) -> _OT: ...
    def pack_single(self, obj: _IT, context: _ContextLike) -> None: ...

class Operator:
    func: Callable[[Any, Any], _StructLike]
    def __init__(self, func: Callable[[Any, Any], _StructLike]) -> None: ...
    def __truediv__(self, arg2) -> _StructLike: ...
    def __rtruediv__(self, arg1) -> Operator: ...
    def __call__(self, arg1, arg2) -> _StructLike: ...

def get_args(
    args: Union[Any, _ContextLambda, List[Union[Any, _ContextLambda]]],
    context: _ContextLike,
) -> list[Any]: ...
def get_kwargs(
    kwargs: Dict[Any, Union[Any, _ContextLambda]], context: _ContextLike
) -> dict: ...



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import itertools

from caterpillar.abc import _PrefixedType
from caterpillar.context import (
    CTX_PATH,
    CTX_FIELD,
    CTX_INDEX,
    CTX_OBJECT,
    CTX_STREAM,
    CTX_SEQ,
    O_CONTEXT_FACTORY,
)
from caterpillar.exception import Stop, StructException, InvalidValueError
from caterpillar.options import O_ARRAY_FACTORY


class WithoutContextVar:
    def __init__(self, context, name, value) -> None:
        self.context = context
        self.old_value = context[name]
        self.value = value
        self.name = name
        self.field = context[CTX_FIELD]

    def __enter__(self) -> None:
        self.context[self.name] = self.value

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.context[self.name] = self.old_value
        # We have to apply the right field as instance of the Field class
        # might set their own value into the context.
        self.context[CTX_FIELD] = self.field


def unpack_seq(context, unpack_one):
    """Generic function to unpack sequenced elements.

    :param stream: the input stream
    :type stream: _StreamType
    :param context: the current context
    :type context: _ContextLike
    :return: a list of unpacked elements
    :rtype: List[Any]
    """
    stream = context[CTX_STREAM]
    field = context[CTX_FIELD]
    assert field and context[CTX_SEQ]

    length = field.length(context)
    base_path = context[CTX_PATH]
    # Special elements '_index' and '_length' can be referenced within
    # the new context. The '_pos' attribute will be adjusted automatically.
    values = []  # always list (maybe add factory)
    seq_context = O_CONTEXT_FACTORY.value(
        _root=context._root,
        _parent=context,
        _io=stream,
        _length=length,
        _lst=values,
        _field=field,
        _obj=context.get(CTX_OBJECT),
        _is_seq=False,
    )
    greedy = length is Ellipsis
    # pylint: disable-next=unidiomatic-typecheck
    prefixed = type(length) is _PrefixedType
    if prefixed:
        # We have to temporarily remove the array status from the parsing field
        with WithoutContextVar(context, CTX_SEQ, False):
            field.amount = 1
            new_length = length.start.__unpack__(context)
            field.amount, length = length, new_length

        if not isinstance(length, int):
            raise InvalidValueError(
                f"Prefix struct returned non-integer: {length!r}", context
            )

    for i in range(length) if not greedy else itertools.count():
        try:
            seq_context[CTX_PATH] = f"{base_path}.{i}"
            seq_context[CTX_INDEX] = i
            values.append(unpack_one(seq_context))

            # NOTE: we introduce this check to reduce time when unpacking
            # a greedy range of elements.
            if greedy and iseof(stream):
                break
        except Stop:
            break
        except Exception as exc:
            if greedy:
                break
            raise StructException(str(exc), context) from exc

    if O_ARRAY_FACTORY.value:
        return O_ARRAY_FACTORY.value(values)
    return values


def pack_seq(seq, context, pack_one) -> None:
    """Generic function to pack sequenced elements.

    :param seq: the iterable of elements
    :type seq: Iterable
    :param stream: the output stream
    :type stream: _StreamType
    :param context: the current operation context
    :type context: _ContextLike
    :raises InvalidValueError: if the input object is not iterable
    """
    stream = context[CTX_STREAM]
    field = context[CTX_FIELD]
    base_path = context[CTX_PATH]
    # REVISIT: when to use field.length(context)
    count = len(seq)
    length = field.amount
    # pylint: disable-next=unidiomatic-typecheck
    if type(length) is _PrefixedType:
        struct = length.start
        # We have to temporatily alter the field's values,
        with WithoutContextVar(context, CTX_SEQ, False):
            field.amount = 1
            struct.__pack__(count, context)
            field.amount = length

    # Special elements '_index' and '_length' can be referenced within
    # the new context. The '_pos' attribute will be adjusted automatically.
    seq_context = O_CONTEXT_FACTORY.value(
        _root=context._root,
        _parent=context,
        _io=stream,
        _length=count,
        _field=field,
        _obj=context.get(CTX_OBJECT),
        # We have to unset the sequence status as we are going to call 'unpack_one'
        _is_seq=False,
    )
    for i, elem in enumerate(seq):
        # The path will contain an additional hint on what element is processed
        # at the moment.
        try:
            seq_context[CTX_INDEX] = i
            seq_context[CTX_PATH] = f"{base_path}.{i}"
            seq_context[CTX_OBJECT] = elem
            pack_one(elem, seq_context)
        except Stop:
            break
        except Exception as exc:
            raise StructException(str(exc), seq_context) from exc


def iseof(stream) -> bool:
    """
    Check if the stream is at the end of the file.

    :param _StreamType stream: The input stream.
    :return: True if the stream is at the end of the file, False otherwise.
    :rtype: bool
    """
    pos = stream.tell()
    eof = not stream.read(1)
    stream.seek(pos)
    return eof



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Any
from io import BytesIO
from caterpillar.abc import (
    _StructLike,
    _GreedyType,
    _PrefixedType,
)
from caterpillar.byteorder import ByteOrder, SysNative, system_arch
from caterpillar.exception import (
    DynamicSizeError,
    StructException,
    OptionError,
    ValidationError,
    InvalidValueError,
)
from caterpillar.options import (
    GLOBAL_FIELD_FLAGS,
    F_DYNAMIC,
    Flag,
)
from caterpillar.context import CTX_OFFSETS, CTX_STREAM, CTX_FIELD, CTX_VALUE, CTX_SEQ
from caterpillar import registry
from caterpillar.shared import getstruct, typeof


def singleton(cls):
    """Simple wrapper to enable singleton structs."""
    return cls()


# Constant representing an invalid default value
INVALID_DEFAULT = object()
DEFAULT_OPTION = object()


class Field:
    """Represents a field in a data structure.

    :param struct: The structure or callable used to define the field's type.
    :param order: Byte order for the field (default: SysNative).
    :param offset: Field offset or callable (default: -1, meaning no explicit offset).
    :param flags: Optional set of flags associated with the field.
    :param amount: The number of elements if this field is a sequence (default: 0).
    :param options: A dictionary representing a switch-case mapping.
    :param condition: A boolean or callable determining whether the field is active.
    :param arch: Architecture specification (default: system_arch).
    :param default: The default value (default: INVALID_DEFAULT).
    :param bits: Bit size if the field is bit-packed.
    """

    def __init__(
        self,
        struct,
        order=None,
        offset=-1,
        flags=None,
        amount=0,
        options=None,
        condition=True,
        arch=None,
        default=INVALID_DEFAULT,
        bits=None,
    ) -> None:
        # Cached flags to optimize packing/unpacking performance
        self._is_lambda = False
        self._has_cond = False
        self._cond_is_lambda = False
        self._is_seq = False
        self._keep_pos = True
        self._has_offset = False
        self._offset_is_lambda = False
        self._amount_is_lambda = False
        self._switch_is_lambda = False
        self._switch_has_default = False

        # private variable initialization
        self.__struct = None
        self.__condition = None
        self.__flags = None
        self.__offset = None
        self.__amount = None
        self.__options = None  # Historically named; represents switch mappings

        # initialization via property setters
        self.struct = struct
        self.condition = condition
        self.order = order or SysNative
        self.flags = set(flags or [])
        self.bits = bits
        self.arch = arch or system_arch
        self.offset = offset
        self.amount = amount or None
        self.options = options

        # INVALID_DEFAULT indicates that no default was explicitly set;
        # allows None to be used as a valid default
        self.default = default

    # -- Property Definitions (Input validation not enforced) --

    @property
    def struct(self):
        """
        The internal structure for this field.
        """
        return self.__struct

    @struct.setter
    def struct(self, value):
        self.__struct = getstruct(value, value)
        # pre-computed state of this field
        self._is_lambda = callable(self.__struct)

    @property
    def condition(self):
        """
        The field's condition expression or value.
        """
        return self.__condition

    @condition.setter
    def condition(self, value):
        self.__condition = value
        self._has_cond = value not in (True, None)
        self._cond_is_lambda = callable(value)

    @property
    def flags(self):
        """The set of flags associated with this field."""
        return self.__flags

    @flags.setter
    def flags(self, value):
        self.__flags = value

    def add_flag(self, flag) -> None:
        """
        Adds a flag to this field.

        :param flag: The flag to add.

        .. versionadded:: 2.6.0
        """
        self.flags.add(flag)

    def has_flag(self, flag: Flag) -> bool:
        """Checks whether this field stores the given flag.

        :param flag: the flag to lookup
        :type flag: Flag
        :return: true if this flag has been found
        :rtype: bool
        """
        return flag in self.flags or flag in GLOBAL_FIELD_FLAGS

    def remove_flag(self, flag: Flag) -> None:
        """
        Removes a flag from this field.

        :param flag: The flag to remove.

        .. versionadded:: 2.6.0
        """
        self.flags.discard(flag)

    @property
    def offset(self):
        """The field's offset."""
        return self.__offset

    @offset.setter
    def offset(self, value):
        self.__offset = value
        self._has_offset = value not in (-1, None)
        self._offset_is_lambda = callable(value)
        self._keep_pos = value in (-1, None)

    @property
    def amount(self):
        """The repetition count for this field."""
        return self.__amount

    @amount.setter
    def amount(self, value):
        self.__amount = value
        self._amount_is_lambda = callable(value)
        self._is_seq = (value not in (0, None)) or self._amount_is_lambda

    @property
    def options(self):
        """The switch-case options dictionary."""
        return self.__options

    @options.setter
    def options(self, value):
        self.__options = value
        self._switch_is_lambda = callable(value)
        self._switch_has_default = (
            value and not self._switch_is_lambda and DEFAULT_OPTION in value
        )

    def _verify_context_value(self, value, expected) -> None:
        """
        Verifies that a value is either of the expected type(s) or is a callable.
        Used to validate inputs for context-aware fields.

        :param value: The value to validate.
        :param expected: A type or tuple of valid types.
        :raises TypeError: If validation fails.
        """
        # As the offset value or amount may be dynamic, we have to candidate
        # types. There should be an error if none applies.
        if not isinstance(value, expected) and not callable(value):
            raise TypeError(
                f"Expected a valid value or context lambda, got {type(value)}"
            )

    # --- Operator Overloads ---

    def __or__(self, flag):  # add flags
        """
        Adds a flag using the '|' operator.

        :param flag: The flag to add.
        :return: The updated Field instance.
        """
        self.add_flag(flag)
        return self

    def __xor__(self, flag):  # remove flags:
        """
        Removes a flag using the '^' operator.

        :param flag: The flag to remove.
        :return: The updated Field instance.
        """
        self.remove_flag(flag)
        return self

    def __matmul__(self, offset):
        """
        Sets the field's offset using the '@' operator.

        :param offset: The offset value (int or callable).
        :return: The updated Field instance.
        """
        self._verify_context_value(offset, int)
        self.offset = offset
        return self

    def __getitem__(self, dim):
        """
        Sets the number of elements using the indexing operator (e.g.,
        field[3]).

        :param dim: Repetition count or dynamic type.
        :return: The updated Field instance.
        """
        self._verify_context_value(dim, (_GreedyType, int, _PrefixedType))
        self.amount = dim
        return self

    def __rshift__(self, switch):
        """
        Defines switch-case mappings using the '>>' operator.

        :param switch: A dictionary of options.
        :return: The updated Field instance.
        """
        self._verify_context_value(switch, dict)
        self.options = switch
        return self

    def __floordiv__(self, condition):
        """
        Sets the field's condition using the '//' operator.

        :param condition: A boolean or callable.
        :return: The updated Field instance.
        """
        self._verify_context_value(condition, bool)
        self.condition = condition
        return self

    def __rsub__(self, bits):
        """
        Sets the number of bits for bit-packed fields using the reverse '-'
        operator.

        :param bits: Bit width (int or callable).
        :return: The updated Field instance.
        """
        self._verify_context_value(bits, int)
        self.bits = bits
        return self

    def __set_byteorder__(self, order: ByteOrder):
        """
        Explicitly sets the byte order for this field.

        :param order: A ByteOrder instance.
        :return: The updated Field instance.
        """
        self.order = order
        return self

    def __type__(self):
        """
        Returns the resolved type of the field, typically from the struct.

        :return: The field type.
        """
        return self.get_type()

    __ixor__ = __xor__
    __ior__ = __or__
    __ifloordiv__ = __floordiv__
    __irshift__ = __rshift__
    __imatmul__ = __matmul__
    __isub__ = __rsub__

    def is_seq(self) -> bool:
        """Returns whether this field is sequential.

        :return: whether this field is sequental
        :rtype: bool
        """
        return self._is_seq

    def is_enabled(self, context) -> bool:
        """Evaluates the condition of this field.

        :param context: the context on which to operate
        :type context: _ContextLike
        :return: ``True``, if this field is enabled
        :rtype: bool
        """
        if not self._has_cond:
            return True

        return self.condition(context) if self._cond_is_lambda else self.condition

    def has_condition(self) -> bool:
        """Returns whether this field is linked to a condition"""
        return self._has_cond

    def length(self, context):
        """Calculates the sequence length of this field.

        :param context: the context on which to operate
        :type context: _ContextLike
        :raises DynamicSizeError: if this field has a dynamic size
        :return: the number of elements
        :rtype: Union[int, _GreedyType]
        """
        try:
            return self.amount(context) if self._amount_is_lambda else self.amount
        except Exception as exc:
            raise DynamicSizeError("Dynamic sized field!", context) from exc

    def get_struct(self, value, context):
        """Returns the struct from stored options.

        :param value: the unpacked or packed value
        :type value
        :param context: the current context
        :type context: _ContextLike
        :return: the struct that packs or unpacks the data
        :rtype: _StructLike
        """
        # treat 'value' as the key of specified options
        if self._switch_is_lambda:
            struct = self.options(value, context)
        else:
            if value not in self.options and not self._switch_has_default:
                raise OptionError(f"Option {value!r} not found!", context)

            struct = self.options.get(value, None) or self.options.get(DEFAULT_OPTION)

        if struct is None:
            # The struct must be non-null
            raise InvalidValueError(
                f"Could not find switch value for: {value!r}", context
            )

        return getstruct(struct, struct)

    def get_offset(self, context) -> int:
        """Returns the offset position of this field"""
        return self.offset(context) if self._offset_is_lambda else self.offset

    def get_type(self):
        """Returns the annotation type for this field

        :return: the annotation type
        :rtype: type
        """
        if not self.options:
            type_ = typeof(self.struct)
            return type_ if not self.is_seq() else list[type_]

        # We construct a Union type hint as an alternative:
        if self._switch_is_lambda:
            return Any

        return_type = Any
        for switch_type in self.options.values():
            return_type = return_type | typeof(switch_type)

        return return_type

    def get_name(self):
        return getattr(self, "__name__", None)

    # IO related stuff
    def __unpack__(self, context):
        """Reads packed data from the given stream.

        This method returns nothing if this field is disabled and applies switch if
        additional options are configured.

        :param context: the current context
        :type context: _ContextLike
        :return: the parsed data
        """
        stream = context[CTX_STREAM]
        if self._has_cond and not self.is_enabled(context):
            # Disabled fields or context lambdas won't pack any data
            return

        # Using this inlined version of self.is_seq(), we reduce the amount of
        # calls made to the method and save A LOT of time.
        # pylint: disable-next=protected-access
        context[CTX_SEQ] = self._is_seq
        # pylint: disable-next=protected-access
        keep_pos = self._keep_pos
        if not self._is_lambda:  # REVISIT: maybe hardcode this
            if not keep_pos:
                fallback = stream.tell()

            if self._has_offset:
                offset = self.offset(context) if self._offset_is_lambda else self.offset
                stream.seek(offset)

            context[CTX_FIELD] = self
            # Switch is applicable AFTER we parsed the first value
            try:
                value = self.struct.__unpack__(context)
                if not keep_pos:
                    stream.seek(fallback)
            # pylint: disable-next=broad-exception-caught
            except Exception as exc:
                if not isinstance(exc, StructException):
                    exc = StructException(str(exc), context)
                # Any exception leads to a default value if configured
                value = self.default
                if value is INVALID_DEFAULT or isinstance(exc, ValidationError):
                    raise exc
        else:
            # Context functions should be executed with top priority
            value = self.struct(context)

        # unpack using switch
        if self.options:
            struct = self.get_struct(value, context)
            # The "keep_position" flag is not applicable here. Configure a field to keep the
            # position afterward.
            context[CTX_VALUE] = value
            value = struct.__unpack__(context)

        return value

    def __pack__(self, obj, context) -> None:
        """Writes the given object to the provided stream.

        There are several options associated with this function. First, disabled
        fields will never pack any data. Next, if an offset position is specified,
        this method will first jump to that position and then parse all data.

        If this field contains a sequential type, it will automatically iterate over
        all elements in ``obj``. The final stream position will be reset if the user
        flag ``KEEP_POSITION`` is not found.

        :param obj: the value to write
        :type obj
        :param context: the current context with a qualified path
        :type context: _ContextLike
        :raises TypeError: if the value is not iterable but this field is marked
                           to be sequential
        """
        # TODO: revisit code
        if self._has_cond and not self.is_enabled(context):
            # Disabled fields or context lambdas won't pack any data
            return

        stream = context[CTX_STREAM]
        keep_pos = self._keep_pos
        has_offset = self._has_offset
        context[CTX_FIELD] = self
        # pylint: disable-next=protected-access
        context[CTX_SEQ] = self._is_seq
        if not keep_pos or has_offset:
            fallback = stream.tell()

        if has_offset:
            # TODO: implement F_OFFSET_OVERRIDE
            # We write the current field into a temporary memory buffer
            # and add it after all processing hasbeen finished.
            offset = self.offset(context) if self._offset_is_lambda else self.offset
            base_stream = stream
            stream = BytesIO()
            context[CTX_STREAM] = stream

        if not self.options:
            _ = (
                self.struct.__pack__(obj, context)
                if not self._is_lambda
                else self.struct(context)
            )
        else:
            # Just hand over the input value if the struct is not a lambda
            context[CTX_VALUE] = obj
            value = self.struct(context) if self._is_lambda else obj
            if not self._is_lambda:
                # support for non-context lambdas with switch statements
                self.struct.__pack__(value, context)
            if self.options is not None:
                struct = self.get_struct(value, context)
                struct.__pack__(obj, context)

        if not has_offset and not keep_pos:
            # The position shouldn't be persisted reset the stream
            stream.seek(fallback)

        if has_offset:
            # Place the stream into the internal offset map
            context._root[CTX_OFFSETS][offset] = stream.getbuffer()
            context[CTX_STREAM] = base_stream

    def __size__(self, context) -> int:
        """Calculates the size of this field.

        There are several situations to bear in mind when executing this function:

         1. If this field is disabled, it will return zero as its size.
         2. Next, dynamic fields do not store a size and DynamicSizeError will be thrown.
         3. Sequential fields will multiply the struct's size with the amount of elements.

         :param context: the operation context
         :type context: _ContextLike
         :raises DynamicSizeError: if this field has a dynamic size
         :return: the calculated size
         :rtype: int
        """
        # 1. If this field is disabled, it will return zero as its size
        if not self.is_enabled(context):
            return 0

        # 2. Next, dynamic fields does not store a size
        if self.has_flag(F_DYNAMIC):
            raise DynamicSizeError("Dynamic sized field!", context)

        context[CTX_FIELD] = self
        context[CTX_SEQ] = self._is_seq

        # 3. We should gather the element count if this field stores
        # a sequential element
        count = 1
        if context[CTX_SEQ]:
            count = self.length(context)
            if isinstance(count, _GreedyType):
                raise DynamicSizeError(
                    "Greedy field does not have a fixed size!", context
                )

        struct = self.struct
        if self.options:
            if not callable(self.struct):
                raise DynamicSizeError(
                    "Switch statement without ContextLambda is danymic sized!", context
                )
            try:
                value = self.struct(context)
            except Exception as exc:
                raise DynamicSizeError("Dynamic sized switch field!", context) from exc
            struct = self.get_struct(value, context)

        size = struct.__size__(context)
        return count * size

    # representation, maybe revisit
    def __str__(self) -> str:
        name = self.get_name() or "<unnamed>"
        offset = f", offset={self.offset}" if self._has_offset else ""
        return (
            f"Field({name!r}, arch={self.arch.name!r}, order={self.order.name!r}, "
            f"seq={self.is_seq()}, struct={self.struct!r}, cond={self.condition is not True}, "
            f"options={bool(self.options)}{offset})"
        )

    def __repr__(self) -> str:
        return self.__str__()


# --- private type converter ---
@registry.TypeConverter(_StructLike)
def _type_converter(annotation, kwargs):
    # REVISIT: more options ?
    arch = kwargs.pop("arch", None)
    order = kwargs.pop("order", None)
    return Field(annotation, order=order, arch=arch)


registry.annotation_registry.append(_type_converter)


class _CallableTypeConverter(registry.TypeConverter):
    def matches(self, annotation) -> bool:
        # must be a callable but not a type
        return callable(annotation) and not isinstance(annotation, type)

    def convert(self, annotation, kwargs):
        arch = kwargs.pop("arch", None)
        order = kwargs.pop("order", None)
        # callables are treates as context lambdas
        return Field(annotation, order=order, arch=arch)


registry.annotation_registry.append(_CallableTypeConverter())



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import sys

class StructException(Exception):
    """Base class for all struct-related exceptions"""

    def __init__(self, message: str, context=None) -> None:
        super().__init__(message)
        self.context = context
        if context and sys.version_info >= (3, 12):
            self.add_note(f"Context-Path: {context.__context_getattr__('_path')}")


class DynamicSizeError(StructException):
    """
    Exception raised for errors related to dynamic sizes in structs.

    This can be used when there are issues with determining the size of a struct element at runtime.
    """


class OptionError(StructException):
    """
    Exception raised for errors related to options in structs.

    This can be used when there are issues with handling options or choices within a struct.
    """


class ValidationError(StructException):
    """
    Exception raised for errors related to validation in structs.

    This can be used when there are issues with validating the structure or content of a struct.
    """


class UnsupportedOperation(StructException):
    """
    Exception raised for unsupported operations in structs.

    This can be used when attempting an operation that is not supported or allowed in the context of structs.
    """


class InvalidValueError(StructException):
    """
    An exception raised for invalid configuration values.
    """


class StreamError(StructException):
    """
    An exception raised for read and write issues.
    """


class DelegationError(StructException):
    """
    An exception used if underlying calls fail.
    """


class Stop(StructException):
    """Stops the current parsing process"""



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import (
    Any,
    Iterable,
    TypeVar,
)
from typing_extensions import Self, override

from caterpillar.abc import (
    _StructLike,
    _ActionLike,
    _ContextLike,
    _OptionLike,
    _ContextLambda,
    _EndianLike,
    _ArchLike,
)
from caterpillar.byteorder import ByteOrder, Arch
from caterpillar.fields._base import Field
from caterpillar.fields._mixin import FieldMixin

_SeqModelT = TypeVar("_SeqModelT", default=dict[str, Any])

class _Member:
    name: str | None
    field: Field[Any, Any] | _ActionLike
    include: bool
    is_action: bool
    action_unpack: _ContextLambda[None]
    action_pack: _ContextLambda[None]

    def __init__(
        self,
        name: str | None,
        field: Field[Any, Any] | _ActionLike,
        include: bool = ...,
        is_action: bool = ...,
    ) -> None: ...
    @override
    def __eq__(self, value: object, /) -> bool: ...

class Sequence(FieldMixin[_SeqModelT, _SeqModelT]):
    model: dict[str, Any]
    fields: list[_Member]
    order: _EndianLike
    arch: _ArchLike
    options: set[_OptionLike]
    field_options: set[_OptionLike]
    is_union: bool
    _member_map_: dict[str, Field]
    def __init__(
        self,
        model: dict[str, Any],
        order: _EndianLike | None = None,
        arch: _ArchLike | None = None,
        options: Iterable[_OptionLike] | None = None,
        field_options: Iterable[_OptionLike] | None = None,
    ) -> None: ...
    def __add__(self, sequence) -> Self: ...
    def __sub__(self, sequence) -> Self: ...
    __iadd__ = __add__
    __isub__ = __sub__
    def has_option(self, option: _OptionLike) -> bool: ...
    def add_field(self, name: str, field: Field, included: bool = False) -> None: ...
    def add_action(self, action: _ActionLike) -> None: ...
    def del_field(self, name: str, field: Field) -> None: ...
    def get_members(self) -> dict[str, Field]: ...
    def unpack_one(self, context: _ContextLike) -> _SeqModelT: ...
    def get_value(self, obj: Any, name: str, field: Field) -> Any: ...
    def pack_one(self, obj: _SeqModelT, context: _ContextLike) -> None: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def __unpack__(self, context: _ContextLike) -> _SeqModelT: ...
    def __type__(self) -> type[_SeqModelT]: ...
    def __pack__(self, obj: _SeqModelT, context: _ContextLike) -> None: ...
    def _set_default(self, name: str, value: Any) -> Any: ...
    def _included(self, name: str, default: Any | None, annotation: Any) -> bool: ...
    def _process_default(
        self,
        name: str,
        annotation: Any,
        had_default: bool = ...,
    ) -> Any: ...
    def _replace_type(self, name: str, type_: type) -> None: ...
    def _remove_from_model(self, name: str) -> None: ...
    def _process_model(self) -> None: ...
    def _prepare_fields(self) -> dict[str, Any]: ...
    def _process_annotation(
        self,
        annotation: Any,
        default: Any,
        order: ByteOrder,
        arch: Arch,
    ) -> _StructLike: ...
    def _process_field(self, name: str, annotation: Any, default: Any) -> Field: ...



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from caterpillar.exception import InvalidValueError, DynamicSizeError, StreamError
from caterpillar.byteorder import LittleEndian, LITTLE_ENDIAN_FMT
from caterpillar.context import CTX_FIELD, CTX_STREAM
from caterpillar.options import Flag

from ._mixin import FieldStruct


VARINT_LSB = Flag("varint.lsb")


class VarInt(FieldStruct):
    """Variable-length integer struct.

    This class implements variable-length unsigned integer for big-endian and
    little-endian encoding. The default behaviour uses big-endian encoding. It
    has an additional flag that can be configured globally or once per field.

    ``VARINT_LSB`` specifies that the last significant byte will use a ``1`` to
    identify the end of the varint. Otherwise, zero will be used (which is the
    default setting). The following configurations are therefore possible:

    >>> field = be + vint; print(pack(1024, field))
    b'\\x88\\x00'
    >>> field = be + vint | VARINT_LSB; print(pack(1024, field))
    b'\\x08\\x80'
    >>> field = le + vint; print(pack(1024, field))
    b'\\x80\\x08'
    >>> field = le + vint | VARINT_LSB; print(pack(1024, field))
    b'\\x00\\x88'
    """

    __slots__ = ("__byteorder__",)

    def __init__(self):
        self.__byteorder__ = LittleEndian

    def __type__(self) -> type:
        return int

    def __size__(self, context) -> int:
        raise DynamicSizeError("VarInt has dynamic size!")

    def bit_config(self, context) -> tuple:
        high_bit = 1 << 7
        low_bit = 0
        if field := context.get(CTX_FIELD):
            if field.has_flag(VARINT_LSB):
                high_bit = 0
                low_bit = 1 << 7
        return high_bit, low_bit

    def pack_single(self, obj, context) -> None:
        """
        Pack a single value into the stream.

        :param obj: The value to pack.
        :param context: The current context.
        """
        if obj is None:
            raise InvalidValueError("NoneType can't be interpreted as a VarInt!")

        if obj < 0:
            raise InvalidValueError("Invalid negative value for VarInt encoding!")

        stream = context[CTX_STREAM]
        field = context.get(CTX_FIELD)
        order = field.order if field else self.__byteorder__
        is_little = order.ch == LITTLE_ENDIAN_FMT

        hb, lb = self.bit_config(context)
        # This implementation is using LittleEndian. Later we can use reverse to
        # apply little endian encoding.
        data = []
        while obj > 0b01111111:
            data.append(obj & 0x7F)
            obj >>= 7
        data.append(obj)

        if not is_little:
            data.reverse()

        for i in range(0, len(data) - 1):
            data[i] |= hb

        data[-1] |= lb
        # Just write all bytes to the stream
        stream.write(bytes(data))

    def unpack_single(self, context):
        """
        Unpack a single value from the stream.

        :param stream: The input stream.
        :param context: The current context.
        :return: The unpacked value.
        """
        stream = context[CTX_STREAM]
        data = []
        _, lb = self.bit_config(context)
        shift = 0
        is_little = context[CTX_FIELD].order == LittleEndian

        while True:
            # Note tha unpack operation here to retrieve one byte only
            try:
                (value,) = stream.read(1)
            except ValueError as exc:
                raise StreamError("Can't read stream!", context) from exc
            data.append(value & 0x7F)
            if value & 0x80 == lb:
                # The "low_byte" here is taken from the field's configuration
                break

        # The following examples shows what to do with the byteorder
        # configuration: num=0xA0B0C0D0
        #                            0     1     2     3
        # b'\xd0\xc0\xb0\xa0' LE : [0xD0, 0xC0, 0xB0, 0xA0]
        # b'\xa0\xb0\xc0\xd0' BE : [0xA0, 0xB0, 0xC0, 0xD0] << reverse that
        if not is_little:
            data.reverse()
        value = 0
        for number in data:
            value |= number << shift
            shift += 7
        return value


vint = VarInt()



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from caterpillar.abc import _ContextLike

class StructException(Exception):
    context: _ContextLike
    def __init__(self, message: str, context: _ContextLike | None = None) -> None: ...

class DynamicSizeError(StructException): ...
class OptionError(StructException): ...
class ValidationError(StructException): ...
class UnsupportedOperation(StructException): ...
class InvalidValueError(StructException): ...
class StreamError(StructException): ...
class DelegationError(StructException): ...
class Stop(StructException): ...



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from .common import Transformer
from caterpillar.abc import (
    _ContainsStruct,
    _ContextLike,
    _StructLike,
    _LengthT,
)
from typing import Protocol, runtime_checkable

@runtime_checkable
class _Compressor(Protocol):
    def compress(self, data: bytes, **kwds) -> bytes: ...
    def decompress(self, data: bytes, **kwds) -> bytes: ...

class Compressed(Transformer[bytes, bytes, bytes, bytes]):
    compressor: _Compressor
    comp_args: dict
    decomp_args: dict
    def __init__(
        self,
        compressor: _Compressor,
        struct: _ContainsStruct[bytes, bytes] | _StructLike[bytes, bytes],
        comp_kwargs: dict | None = None,
        decomp_kwargs: dict | None = None,
    ) -> None: ...
    def encode(self, obj: bytes, context: _ContextLike) -> bytes: ...
    def decode(self, parsed: bytes, context: _ContextLike) -> bytes: ...

_LengthTorStructT = _LengthT | _ContainsStruct[bytes, bytes] | _StructLike[bytes, bytes]

def compressed(
    lib: _Compressor,
    obj: _LengthTorStructT,
    comp_kwargs: dict | None = None,
    decomp_kwargs: dict | None = None,
) -> _StructLike[bytes, bytes]: ...

def ZLibCompressed(
    obj: _LengthTorStructT,
    comp_kwargs: dict | None = None,
    decomp_kwargs: dict | None = None,
) -> _StructLike[bytes, bytes]: ...
def Bz2Compressed(
    obj: _LengthTorStructT,
    comp_kwargs: dict | None = None,
    decomp_kwargs: dict | None = None,
) -> _StructLike[bytes, bytes]: ...
def LZMACompressed(
    obj: _LengthTorStructT,
    comp_kwargs: dict | None = None,
    decomp_kwargs: dict | None = None,
) -> _StructLike[bytes, bytes]: ...
def LZOCompressed(
    obj: _LengthTorStructT,
    comp_kwargs: dict | None = None,
    decomp_kwargs: dict | None = None,
) -> _StructLike[bytes, bytes]: ...



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from caterpillar.abc import _ContextLambda, _StructLike, _ContainsStruct
from typing import Any

MODE_PACK: int
MODE_UNPACK: int
ATTR_STRUCT: str
ATTR_TYPE: str
ATTR_BYTEORDER: str
ATTR_BITS: str
ATTR_SIGNED: str
ATTR_TEMPLATE: str
ATTR_ACTION_PACK: str
ATTR_ACTION_UNPACK: str

class Action:
    __action_pack__: _ContextLambda
    __action_unpack__: _ContextLambda
    def __init__(
        self,
        pack: _ContextLambda | None = None,
        unpack: _ContextLambda | None = None,
        both: _ContextLambda | None = None,
    ) -> None: ...
    @staticmethod
    def is_action(obj: Any) -> bool: ...

def hasstruct(obj: Any) -> bool: ...
def getstruct(
    obj: Any, /, __default: _StructLike[Any, Any] | None = None
) -> _StructLike[Any, Any] | None: ...
def typeof(struct: _StructLike[Any, Any] | _ContainsStruct[Any, Any]) -> type: ...



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from enum import Enum
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    List,
    Optional,
    Type,
    TypeVar,
    overload,
)

from caterpillar.abc import _ContextLike, _ActionLike
from caterpillar.byteorder import Arch, ByteOrder
from caterpillar.options import B_GROUP_END, B_GROUP_NEW, Flag
from caterpillar.fields._base import Field
from caterpillar.model._struct import Struct

_ModelT = TypeVar("_ModelT")

DEFAULT_ALIGNMENT: int

def getbits(obj: Any) -> int: ...
def issigned(obj: Any) -> bool: ...

class SetAlignment:
    alignment: int
    def __init__(self, new_alignment: int) -> None: ...
    @staticmethod
    def flag(new_alignment: int) -> Flag[int]: ...
    def __hash__(self) -> int: ...

NewGroup = B_GROUP_NEW
EndGroup = B_GROUP_END

_VT = TypeVar("_VT", default=int)

class BitfieldValueFactory(Generic[_VT]):
    def to_int(self, obj: _VT) -> int: ...
    def from_int(self, value: int) -> _VT: ...

_EnumT = TypeVar("_EnumT", bound=Enum)

class EnumFactory(Generic[_EnumT], BitfieldValueFactory[_EnumT | int]):
    strict: bool
    def __init__(self, model: Type[_EnumT], strict: bool = ...) -> None: ...
    def from_int(self, value: int) -> _EnumT | int: ...

class CharFactory(BitfieldValueFactory[str]):
    def from_int(self, value: int) -> str: ...
    def to_int(self, obj: str) -> int: ...

class BitfieldEntry:
    bit: int
    width: int
    name: str
    factory: Optional[BitfieldValueFactory]
    action: Optional[_ActionLike]
    low_mask: int

    def __init__(
        self,
        bit: int,
        width: int,
        name: str,
        factory: BitfieldValueFactory | Type | None = ...,
        action: _ActionLike | None = ...,
    ) -> None: ...
    @staticmethod
    def new_action(action: _ActionLike) -> BitfieldEntry: ...
    def shift(self, value_width: int) -> int: ...
    def is_action(self) -> bool: ...

class BitfieldGroup:
    bit_count: int
    entries: List[BitfieldEntry | Field]
    def __init__(self, bit_count: int) -> None: ...
    def is_field(self) -> bool: ...
    def get_field(self) -> Field: ...
    def set_field(self, field: Field) -> None: ...
    def align_to(self, alignment: int) -> None: ...
    def is_empty(self) -> bool: ...
    def get_size(self, context: _ContextLike | None = ...) -> int: ...
    def get_bits(self, context: _ContextLike | None = ...) -> int: ...

class Bitfield(Struct[_ModelT]):
    alignment: int
    groups: List[BitfieldGroup]
    def __init__(
        self,
        model,
        order: ByteOrder | None = ...,
        arch: Arch | None = ...,
        options: Iterable[Flag] | None = ...,
        field_options: Iterable[Flag] | None = ...,
        alignment: int | None = ...,
    ) -> None: ...
    def __add__(self, sequence): ...
    def __size__(self, context: _ContextLike) -> int: ...
    def __bits__(self) -> int: ...
    def unpack_one(self, context: _ContextLike): ...
    def pack_one(self, obj: _ModelT, context: _ContextLike) -> None: ...
    def add_action(self, action: _ActionLike) -> None: ...
    def get_entry(self, name: str) -> Optional[BitfieldEntry]: ...
    def _process_align(
        self, options: Optional[Iterable[Flag | SetAlignment]]
    ) -> Field: ...
    def _process_bits(
        self,
        name: str,
        bits: int,
        factory: Optional[BitfieldValueFactory | Type] = ...,
        options: Optional[Iterable[Flag | SetAlignment]] = ...,
    ) -> Field: ...
    def _process_bits_field(
        self,
        name: str,
        field: Field,
        factory: Optional[BitfieldValueFactory | Type] = ...,
        options: Optional[Iterable[Flag | SetAlignment]] = ...,
    ) -> Field: ...
    def _process_options(
        self,
        options: Iterable[Flag | SetAlignment],
        entry: Optional[BitfieldEntry] = ...,
    ) -> bool: ...
    def _new_group(self, alignment: int) -> BitfieldGroup: ...
    def _process_alignment_option(self, option: Flag | SetAlignment) -> bool: ...

@overload
def bitfield(
    cls: None = None,
    /,
    *,
    options: Iterable[Flag] | None = ...,
    order: ByteOrder | None = ...,
    arch: Arch | None = ...,
    field_options: Iterable[Flag] | None = ...,
) -> Callable[[Type[_ModelT]], Type[_ModelT]]: ...
@overload
def bitfield(
    cls: Type[_ModelT],
    /,
    *,
    options: Iterable[Flag] | None = ...,
    order: ByteOrder | None = ...,
    arch: Arch | None = ...,
    field_options: Iterable[Flag] | None = ...,
) -> Type[_ModelT]: ...



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Protocol, runtime_checkable

from caterpillar.exception import UnsupportedOperation
from caterpillar.exception import InvalidValueError
from caterpillar.context import CTX_STREAM
from .common import Memory, Bytes
from ._mixin import get_args, get_kwargs


@runtime_checkable
class Padding(Protocol):  # pylint: disable=missing-class-docstring
    def unpadder(self):
        """Abstract method to get an unpadder for padding."""

    def padder(self):
        """Abstract method to get a padder for padding."""


class Encrypted(Memory):
    """Struct that is able to encrypt/decrypt blocks of memory.

    :param length: Length of the encrypted data.
    :type length: Union[int, _GreedyType, _ContextLambda]
    :param algorithm: Encryption algorithm.
    :type algorithm: Type[algorithms.CipherAlgorithm]
    :param mode: Encryption mode.
    :type mode: Union[Type[modes.Mode], modes.Mode]
    :param padding: Padding scheme for encryption.
    :type padding: Union[Padding, Type[Padding]], optional
    :param algo_args: Additional arguments for the encryption algorithm.
    :type algo_args: Optional[Iterable[_ArgType]], optional
    :param mode_args: Additional arguments for the encryption mode.
    :type mode_args: Optional[Iterable[_ArgType]], optional
    :param padding_args: Additional arguments for the padding scheme.
    :type padding_args: Optional[Iterable[_ArgType]], optional
    :param post: Post-processing structure.
    """

    # REVISIT: this constructor looks ugly
    def __init__(
        self,
        length,
        algorithm,
        mode,
        padding=None,
        algo_args=None,
        mode_args=None,
        padding_args=None,
        post=None,
    ) -> None:
        try:
            from cryptography.hazmat.primitives.ciphers import Cipher
        except ImportError:
            raise UnsupportedOperation(
                (
                    "To use encryption with this framework, the module 'cryptography' "
                    "is required! You can install it via pip or use the packaging "
                    "extra 'crypto' that is available with this library."
                )
            )

        super().__init__(length)
        self._algo = algorithm
        self._algo_args = algo_args
        self._mode = mode
        self._mode_args = mode_args
        self._padding = padding
        self._padding_args = padding_args
        self.post = post

    def algorithm(self, context):
        """
        Get the encryption algorithm instance.

        :param context: The current operation context.
        :type context: _ContextLike
        :return: An instance of the encryption algorithm.
        :rtype: algorithms.CipherAlgorithm
        """
        from cryptography.hazmat.primitives.ciphers import CipherAlgorithm

        return self.get_instance(CipherAlgorithm, self._algo, self._algo_args, context)

    def mode(self, context):
        """
        Get the encryption mode instance.

        :param context: The current operation context.
        :type context: _ContextLike
        :return: An instance of the encryption mode.
        :rtype: modes.Mode
        """
        from cryptography.hazmat.primitives.ciphers import modes

        return self.get_instance(modes.Mode, self._mode, self._mode_args, context)

    def padding(self, context) -> Padding:
        """
        Get the padding scheme instance.

        :param context: The current operation context.
        :type context: _ContextLike
        :return: An instance of the padding scheme.
        :rtype: Padding
        """
        return self.get_instance(Padding, self._padding, self._padding_args, context)

    def get_instance(self, type_, field, args, context):
        """
        Get an instance of a specified type.

        :param type_: The desired type of the instance.
        :type type_: type
        :param field: The field or instance.
        :type field: Any
        :param args: Additional arguments for the instance.
        :type args: Any
        :param context: The current operation context.
        :type context: _ContextLambda
        :return: An instance of the specified type.
        :rtype: Any
        """
        if isinstance(field, type_) or not field:
            return field

        if isinstance(args, dict):
            args, kwargs = (), get_kwargs(args, context)
        else:
            args, kwargs = get_args(args, context), {}
        return field(*args, **kwargs)

    def pack_single(self, obj, context) -> None:
        """
        Pack a single element.

        :param obj: The element to pack.
        :type obj: Any
        :param context: The current operation context.
        :type context: _ContextLike
        """
        from cryptography.hazmat.primitives.ciphers import Cipher

        cipher = Cipher(self.algorithm(context), self.mode(context))
        padding = self.padding(context)

        data = obj if isinstance(obj, bytes) else bytes(obj)
        if padding:
            padder = padding.padder()
            data = padder.update(data) + padder.finalize()

        encryptor = cipher.encryptor()
        super().pack_single(encryptor.update(data) + encryptor.finalize(), context)

    def unpack_single(self, context):
        """
        Unpack a single element.

        :param context: The current operation context.
        :type context: _ContextLike
        :return: The unpacked element as a memoryview.
        :rtype: memoryview
        """
        from cryptography.hazmat.primitives.ciphers import Cipher
        value = super().unpack_single(context)
        cipher = Cipher(self.algorithm(context), self.mode(context))

        decryptor = cipher.decryptor()
        data = decryptor.update(bytes(value)) + decryptor.finalize()

        padding = self.padding(context)
        if padding:
            unpadder = padding.unpadder()
            data = unpadder.update(data) + unpadder.finalize()
        return memoryview(data)


class KeyCipher(Bytes):
    # key: bytes
    # """The key that should be applied.

    # It will be converted automatically to bytes if not given.
    # """

    # key_length: int
    # """Internal attribute to keep track of the key's length"""

    __slots__ = "key", "key_length", "is_lazy"

    def __init__(self, key, length=None) -> None:
        super().__init__(length or ...)
        self.key = self.is_lazy = self.key_length = None
        self.set_key(key)

    def set_key(self, key, context=None) -> None:
        if callable(key) and context is None:
            # context lambda indicates the key will be computed at runtime
            self.key = key
            self.key_length = -1
            self.is_lazy = True
            return

        match key:
            case str():
                self.key = key.encode()
            case int():
                self.key = bytes([key])
            case bytes():
                self.key = key
            case _:
                raise InvalidValueError(
                    f"Expected a valid key type, got {key!r}", context
                )

        self.key_length = len(self.key)
        self.is_lazy = False

    def process(self, obj: bytes, context) -> bytes:
        length = len(obj)
        data = bytearray(length)
        key = self.key
        if self.is_lazy:
            self.set_key(key(context), context)

        self._do_process(obj, data)
        return bytes(data)

    def _do_process(self, src: bytes, dest: bytearray):
        raise NotImplementedError

    def pack_single(self, obj: bytes, context) -> None:
        context[CTX_STREAM].write(self.process(obj, context))

    def unpack_single(self, context) -> bytes:
        obj: bytes = super().unpack_single(context)
        return self.process(obj, context)


class Xor(KeyCipher):
    __slots__ = ()

    def _do_process(self, src: bytes, dest: bytearray):
        for i, e in enumerate(src):
            dest[i] = e ^ self.key[i % self.key_length]


class Or(KeyCipher):
    __slots__ = ()

    def _do_process(self, src: bytes, dest: bytearray):
        for i, e in enumerate(src):
            dest[i] = e | self.key[i % self.key_length]


class And(KeyCipher):
    __slots__ = ()

    def _do_process(self, src: bytes, dest: bytearray):
        for i, e in enumerate(src):
            dest[i] = e & self.key[i % self.key_length]



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import struct as PyStruct
import warnings

from io import BytesIO
from typing import Any
from typing_extensions import Self
from types import NoneType
from functools import cached_property
from enum import Enum as _EnumType
from uuid import UUID

from caterpillar.abc import _StructLike, _StreamType, _ContextLike, _PrefixedType
from caterpillar.exception import (
    ValidationError,
    InvalidValueError,
    DynamicSizeError,
)
from caterpillar.context import CTX_FIELD, CTX_STREAM, CTX_SEQ
from caterpillar.options import Flag, GLOBAL_FIELD_FLAGS
from caterpillar.byteorder import LITTLE_ENDIAN_FMT, LittleEndian, SysNative
from caterpillar import registry
from caterpillar._common import WithoutContextVar
from caterpillar.shared import getstruct

from ._base import Field, INVALID_DEFAULT, singleton
from ._mixin import FieldStruct

# Explicitly report deprecation warnings
warnings.filterwarnings("default", category=DeprecationWarning, module=__name__)

ENUM_STRICT = Flag("enum.strict")


class PyStructFormattedField(FieldStruct):
    """
    A field class representing a binary format using format characters (e.g., 'i', 'x', etc.).

    This class allows you to define fields for binary struct formats using format characters,
    which are often used in Python's `struct` module. Each field is associated with a format character
    (like 'x' for padding or 'i' for a signed integer), and it determines how data is packed and unpacked.

    Example usage:

    >>> field = PyStructFormattedField('i', int)
    >>> field.__repr__()
    "<int32>"

    :param ch: The format character (e.g., 'i', 'x', 'f') that defines how the field is packed and unpacked.
    :param type_: The Python type that corresponds to the format character.
    """

    __slots__ = {
        "text": """The format char (e.g. 'x', 'i', ...)""",
        "ty": """The returned Python type""",
        "_is_padding": """Internal field that indicates this struct is padding""",
    }

    def __init__(self, ch: str, type_: type) -> None:
        self.text = ch
        self.ty = type_
        self.__bits__ = PyStruct.calcsize(self.text) * 8
        self._is_padding = self.text == "x"
        self.__byteorder__ = SysNative

    def __fmt__(self) -> str:
        """
        Get the format character for this field.

        :return: The format character (e.g., 'i', 'x').
        :rtype: str
        """
        return self.text

    def __repr__(self) -> str:
        """
        String representation of the FormatField.

        :return: A string representation.
        """
        if self._is_padding:
            return "<padding>"

        return f"<{self.ty.__name__}{self.__bits__}>"

    def __type__(self) -> type:
        """
        Get the Python type associated with the format specifier.

        :return: The Python type (e.g., `int`, `float`, etc.).
        :rtype: type
        """
        return self.ty

    def __size__(self, context) -> int:
        """
        Calculate the size of the field in bytes.

        The size is calculated by dividing the bit size by 8 (since 1 byte = 8 bits).

        :param context: The context for the current field (can be used for calculating dynamic sizes).
        :return: The size of the field in bytes.
        :rtype: int
        """
        return self.__bits__ // 8

    def pack_single(self, obj, context) -> None:
        """
        Pack a single value into the stream using the defined format character.

        :param obj: The value to pack (can be of the type defined by the format character).
        :param context: The context that provides the stream and field-specific information.
        """
        if obj is None and not self._is_padding:
            return  # Skip packing if the value is None and the field is not padding

        field = context.get(CTX_FIELD)
        order_ch = field.order.ch if field else self.__byteorder__.ch
        fmt = f"{order_ch}{self.text}"
        data = PyStruct.pack(fmt) if self._is_padding else PyStruct.pack(fmt, obj)
        context[CTX_STREAM].write(data)

    def pack_seq(self, seq, context) -> None:
        """
        Pack a sequence of values into the stream.

        If the context specifies a fixed amount, it will pack the sequence accordingly.

        :param seq: The sequence of values to pack.
        :param context: The context providing packing information.
        """
        field = context.get(CTX_FIELD)
        target_length = len(seq)
        if target_length == 0:
            return  # nothing to do

        if not field:
            # just pack directly
            # WE LOSE SIZE CHECKING HERE!
            fmt = f"{self.__byteorder__.ch}{target_length}{self.text}"
        else:
            length = field.length(context)
            if type(length) is _PrefixedType:
                context[CTX_SEQ] = False
                length.start.__pack__(len(seq), context)
                context[CTX_SEQ] = True
            elif length is not Ellipsis:
                if length != target_length:
                    raise ValueError(
                        f"Length mismatch: expected {length}, but only "
                        + f"{target_length} elements were provided!"
                    )

            fmt = f"{field.order.ch}{target_length}{self.text}"

        context[CTX_STREAM].write(PyStruct.pack(fmt, *seq))

    def unpack_single(self, context):
        """
        Unpack a single value from the stream.

        :param context: The context that provides the stream and field-specific information.
        :return: The unpacked value, converted to the field's corresponding Python type.
        """
        field = context.get(CTX_FIELD)
        order_ch = field.order.ch if field else self.__byteorder__.ch
        fmt = f"{order_ch}{self.text}"
        size = self.__bits__ // 8
        (value,) = PyStruct.unpack(fmt, context[CTX_STREAM].read(size))
        return value

    def unpack_seq(self, context):
        """
        Unpack a sequence of values from the stream.

        :param context: The context that provides the stream and field-specific information.
        :return: A list of unpacked values.
        """
        # only possible when a Field has been configured
        field = context[CTX_FIELD]
        length = field.length(context)
        if length == 0:
            return []  # maybe add factory

        if length is Ellipsis:
            return super().unpack_seq(context)

        fmt = f"{field.order.ch}{length}{self.text}"
        size = (self.__bits__ // 8) * length
        return list(PyStruct.unpack(fmt, context[CTX_STREAM].read(size)))

    def is_padding(self) -> bool:
        """
        Check if the field represents padding.

        :return: True if the field is padding (i.e., 'x' format character), False otherwise.
        :rtype: bool
        """
        return self._is_padding


# Instances of FormatField with specific format specifiers
padding = PyStructFormattedField("x", NoneType)
char = PyStructFormattedField("c", str)
boolean = PyStructFormattedField("?", bool)

int8 = PyStructFormattedField("b", int)
uint8 = PyStructFormattedField("B", int)
int16 = PyStructFormattedField("h", int)
uint16 = PyStructFormattedField("H", int)
int32 = PyStructFormattedField("i", int)
uint32 = PyStructFormattedField("I", int)
int64 = PyStructFormattedField("q", int)
uint64 = PyStructFormattedField("Q", int)


ssize_t = PyStructFormattedField("n", int)
size_t = PyStructFormattedField("N", int)

float16 = PyStructFormattedField("e", float)
float32 = PyStructFormattedField("f", float)
float64 = PyStructFormattedField("d", float)
double = float64

void_ptr = PyStructFormattedField("P", int)


class Transformer(FieldStruct):
    """
    A class that acts as a transformer for encoding and decoding data using a wrapped _StructLike object.
    """

    __slots__ = ("struct",)

    def __init__(self, struct) -> None:
        self.struct = struct
        self.__bits__ = getattr(self.struct, "__bits__", None)

    def __type__(self):
        """
        Get the type of the data encoded/decoded by the transformer.

        :return: The type of the data.
        """
        return self.struct.__type__()

    def __size__(self, context) -> int:
        """
        Get the size of the data encoded/decoded by the transformer.

        :param context: The current context.
        :return: The size of the data.
        """
        return self.struct.__size__(context)

    def encode(self, obj, context):
        """
        Encode data using the wrapped _StructLike object.

        :param obj: The original data to be encoded.
        :param context: The current context.
        :return: The encoded data.
        """
        return obj

    def decode(self, parsed, context):
        """
        Decode data using the wrapped _StructLike object.

        :param parsed: The parsed data to be decoded.
        :param context: The current context.
        :return: The decoded data.
        """
        return parsed

    def pack_single(self, obj, context) -> None:
        """
        Pack a single value into the stream using encoding.

        :param obj: The original data to be encoded and packed.
        :param context: The current context.
        """
        value = self.encode(obj, context)
        self.struct.__pack__(value, context)

    def unpack_single(self, context):
        """
        Unpack a single value from the stream and decode it.

        :param context: The current context.
        :return: The decoded data.
        """
        value = self.struct.__unpack__(context)
        return self.decode(value, context)


class Const(Transformer):
    """
    A specialized Transformer that enforces a constant value during
    encoding and decoding.

    This class ensures that the encoded value is always the same
    constant, and when decoding, it checks if the value matches the
    expected constant. If the value doesn't match, a `ValidationError`
    is raised.

    Example usage:

    >>> # Define a constant value to enforce during encoding/decoding
    >>> const_value = 42
    >>> struct = SomeStruct() # must parse 'const_value`
    >>> field = Const(const_value, struct)
    >>> pack(None, field)
    b"\\x2a"
    >>> unpack(field, b"\\x2a")
    42

    :param value: The constant value to be enforced during encoding and decoding.
    :param struct: The _StructLike object to be wrapped.
    """

    __slots__ = ("value",)

    def __init__(self, value, struct) -> None:
        super().__init__(struct)
        self.value = value

    def encode(self, obj, context):
        """
        Encode data using the constant value. This method will always return
        the constant value, regardless of the input. Therefore, :code:`None`
        can be passed as the `obj` parameter.

        :param obj: The original data to be encoded (ignored in this transformer).
        :param context: The current context (optional, not used in this implementation).
        :return: The constant value.

        Example:
            >>> constant_value = 42
            >>> transformer = Const(constant_value, SomeStruct())
            >>> transformer.encode(None, context) # context is optional
            42
        """
        return self.value

    def decode(self, parsed, context):
        """
        Decode data and ensure it matches the constant value. If the
        parsed value doesn't match, a `ValidationError` is raised.

        :param parsed: The parsed data to be decoded (must match the constant value).
        :param context: The current context (optional, not used in this implementation).
        :return: The constant value if parsed matches the expected constant.
        :raises ValidationError: If the parsed value does not match the constant value.

        Example:
        >>> constant_value = 42
        >>> field = Const(constant_value, SomeStruct())
        >>> unpack(field, b"\\x2a")
        42
        >>> unpack(field, b"\\x24")
        Traceback (most recent call last):
        ...
        ValidationError: Expected 42, got 36
        """
        if parsed != self.value:
            raise ValidationError(f"Expected {self.value!r}, got {parsed!r}", context)
        return self.value


class Enum(Transformer):
    """
    A specialized Transformer for encoding and decoding enumeration values.

    This class is used for transforming between enumeration values and their corresponding
    encoded representations (typically integers). It provides encoding and decoding methods
    to ensure that enum values are serialized and deserialized correctly. Additionally, it supports
    a default value in case an invalid or unrecognized value is encountered during decoding.

    Example usage:

    >>> from enum import Enum as PyEnum
    >>> class Color(PyEnum):
    ...     RED = 1
    ...     GREEN = 2
    ...     BLUE = 3
    ...
    >>> cp_enum = Enum(Color, uint8)
    >>> pack(Color.RED, cp_enum, as_field=True)
    b"\\x01"
    >>> unpack(cp_enum, b"\\x01", as_field=True)
    Color.RED

    :param model: The enumeration model (an object with `_member_map_` and `_value2member_map_` attributes).
    :param struct: The _StructLike object to be wrapped.
    :param default: The default value to return if decoding encounters an unrecognized value.
                    Default is `INVALID_DEFAULT`.
    """

    __slots__ = ("model", "default")

    def __init__(
        self,
        model,
        struct,
        default=INVALID_DEFAULT,
    ) -> None:
        super().__init__(struct)
        self.model = model
        self.default = default

    def __type__(self):
        """
        Determine the type for this transformation, which is either the enum type
        or a union of the enum and struct types, depending on the global field flags.

        :return: The type (either the enum model or a union of enum and struct types).
        """
        # pylint: disable-next=protected-access
        if ENUM_STRICT in GLOBAL_FIELD_FLAGS:
            return self.model

        return self.model | self.struct.__type__()

    def encode(self, obj, context):
        """
        Encode an enumeration value into its corresponding encoded representation.

        :param obj: The original enumeration value to encode (e.g., `Color.RED`).
        :param context: The current context.
        :return: The encoded value (usually an integer representing the enum).
        :raises ValidationError: If the input value is not a valid enum type.

        Example:
        >>> cp_enum.encode(Color.GREEN, context)
        2 # (the integer value of Color.GREEN)
        """
        if not isinstance(obj, _EnumType):
            if field := context.get(CTX_FIELD):
                if field.has_flag(ENUM_STRICT):
                    raise ValidationError(
                        f"Expected enum type, got {type(obj)}", context
                    )
            return obj

        return obj.value

    def decode(self, parsed, context):
        """
        Decode an encoded value (typically an integer) back to its corresponding
        enumeration value.

        :param parsed: The parsed value (usually an integer).
        :param context: The current context.
        :return: The corresponding enumeration value.
        :raises InvalidValueError: If the parsed value cannot be mapped to a valid enum.

        Example:
        >>> cp_enum.decode(1, context)
        Color.RED
        >>> unpack(1, cp_enum, as_field=True)
        Color.RED
        """
        # pylint: disable-next=protected-access
        by_name = self.model._member_map_.get(parsed)
        if by_name is not None:
            return by_name

        # pylint: disable-next=protected-access
        by_value = self.model._value2member_map_.get(parsed)
        if by_value is not None:
            return by_value

        default = self.default
        field = context.get(CTX_FIELD)
        if default is INVALID_DEFAULT and field:
            default = field.default

        if default is INVALID_DEFAULT:
            if field and field.has_flag(ENUM_STRICT):
                raise InvalidValueError(
                    f"Could not find enum for value {parsed!r}", context
                )
            return parsed
        return default


class _EnumTypeConverter(registry.TypeConverter):
    def matches(self, annotation) -> bool:
        return isinstance(annotation, type) and issubclass(annotation, _EnumType)

    def convert(self, annotation, kwargs: dict):
        struct_obj = getstruct(annotation)
        if not struct_obj:
            raise ValidationError(
                f"Could not infer Enum struct: could not find __struct__ attribute on annotation {annotation!r}"
            )
        return Enum(annotation, struct_obj)


# TODO: document why this has to be at first position
registry.annotation_registry.insert(0, _EnumTypeConverter())


class Memory(FieldStruct):
    """
    A class representing a memory field that handles packing and unpacking byte-like objects
    of a specified length.

    The `Memory` class is used to manage fields that contain raw byte data. It allows you to
    pack a byte object (either `bytes` or `memoryview`) into a stream and unpack byte data
    from a stream into a `memoryview`. The length of the memory field can be fixed or dynamically
    determined based on the context. If the length is unspecified, the entire available data in
    the stream is read.

    This class can handle the following scenarios for the `length` argument:
    - **Fixed length**: A specific number of bytes (e.g., `10`).
    - **Dynamic length**: A callable that returns the length based on the current context (e.g., `lambda ctx: ctx["length"]`).
    - **Greedy length**: Using `Ellipsis` (`...`), meaning the entire stream is read regardless of the size.

    Examples:

    >>> # Memory with a fixed length of 10 bytes
    >>> memory = Memory(10)
    >>> memory_dynamic = Memory(lambda ctx: ctx._root.length) # dynamic length based on context
    >>> memory_greedy = Memory(...)
    >>> pack(b"1234567890", memory, as_field=True)
    b"1234567890"
    >>> pack(b"1234567890", memory_dynamic, as_field=True, length=10)
    b"1234567890"
    >>> unpack(b"1234567890", memory_greedy, as_field=True)
    b"1234567890"

    :param length: The length of the memory to pack or unpack. It can be:
                   - An integer specifying the number of bytes to handle.
                   - A callable that dynamically returns the number of bytes based on the context.
                   - Ellipsis (`...`), indicating the length is unspecified and the entire stream should be read.
    """

    __slots__ = ("length",)

    def __init__(
        self,
        length,
    ) -> None:
        self.length = length

    def __type__(self) -> type:
        """
        Return the type of the field, which is `memoryview` for this class.

        :return: `memoryview`
        """
        return memoryview

    def __size__(self, context) -> int:
        """
        Calculate the size of the memory field based on the `length` parameter.

        If the `length` is callable, it will be invoked with the context to get the size.
        If the `length` is `Ellipsis`, the size is unspecified and the entire stream will be read.

        :param context: The current context.
        :return: The size of the field in bytes, or `Ellipsis` if the length is unspecified.
        """
        return self.length(context) if callable(self.length) else self.length

    def pack_single(self, obj, context) -> None:
        """
        Pack a single byte object (memoryview or bytes) into the stream.

        This method writes a byte object (`memoryview` or `bytes`) into the stream,
        ensuring that the size of the object matches the expected length defined by
        the `Memory` field. If the expected size is fixed (not `Ellipsis`), a
        validation is performed to ensure that the length of the `obj` matches
        the expected size.

        Examples:

        >>> memory = Memory(10)
        >>> pack(b"1234567890", memory, as_field=True) # Packs exactly 10 bytes into the stream
        b"1234567890"
        >>> pack(b"1", memory, as_field=True)
        Traceback (most recent call last):
        ...
        ValidationError: Memory field expected 10 bytes, but got 1 bytes instead

        :param obj: The byte object to be packed. This can be either a `memoryview` or a `bytes` object.
        :param context: The current context, which provides access to the stream and any additional metadata needed
                        for packing.

        :raises ValidationError: If the length of the `obj` does not match the expected size.
        """
        size = self.__size__(context)
        if size is not Ellipsis:
            if len(obj) != size:
                raise ValidationError(
                    f"Memory field expected {size} bytes, but got {len(obj)} bytes instead",
                    context,
                )
        context[CTX_STREAM].write(obj)

    def unpack_single(self, context):
        """
        Unpack a single byte object (memoryview) from the stream.

        This method reads bytes from the stream into a `memoryview`. The size
        of the object to be unpacked is determined by the `Memory` field's length.
        If the length is fixed (not `Ellipsis`), it reads exactly that number of
        bytes. If the length is unspecified (i.e., `Ellipsis`), it reads until the
        end of the stream.

        :param context: The current context, which provides access to the stream and
                        any additional metadata needed for unpacking.
        :return: A `memoryview` object representing the unpacked byte data.
        """
        stream: _StreamType = context[CTX_STREAM]
        size = self.__size__(context)
        return memoryview(stream.read(size) if size is not Ellipsis else stream.read())


class Bytes(Memory):
    """Byte sequences.

    Same class as :class:`Memory` but with a type of `bytes` instead of `memoryview`.
    """

    __slots__ = ()

    def __type__(self) -> type:
        """
        Return the type associated with this `Bytes` field, which is `bytes`.

        :return: The type (bytes).
        """
        return bytes

    def unpack_single(self, context):
        """
        Unpack a single byte sequence (bytes) from the stream.

        :param context: The current context, which provides access to the stream and
                        other necessary metadata for unpacking.
        :return: A `bytes` object representing the unpacked data from the stream.

        """
        # as we read bytes from the stream, this will return a bytes object
        return super().unpack_single(context).obj


class String(Memory):
    """String sequences.

    Same class as :class:`Memory` but with a type of `str` instead of `memoryview`.

    :param length: The length of the string to pack or unpack. It can be:
                   - An integer specifying the number of bytes (e.g., `10`).
                   - A callable that dynamically returns the number of bytes based on the context (e.g., `lambda ctx: ctx["length"]`).
                   - Ellipsis (`...`), meaning the entire available stream should be read, regardless of size.
    :param encoding: The encoding to use for converting bytes to a string during unpacking. If not specified, defaults to `None`.
                     If specified, the string will be decoded using this encoding (e.g., `utf-8`).
    """

    __slots__ = ("encoding",)

    def __init__(
        self,
        length,
        encoding=None,
    ) -> None:
        super().__init__(length)
        self.encoding = encoding or "utf-8"

    def __type__(self) -> type:
        """
        Return the type associated with this `String` field.

        :return: `str` - the type associated with the `String` field.
        """
        return str

    def pack_single(self, obj: str, context) -> None:
        """Packs a single string into the stream."""
        return super().pack_single(obj.encode(self.encoding), context)

    def unpack_single(self, context):
        """
        Unpack a single string from the stream.

        :param context: The current context.
        :return: A `str` representing the unpacked string, decoded from bytes using the specified encoding.
        """
        return super().unpack_single(context).obj.decode(self.encoding)


class CString(FieldStruct):
    """
    C-style strings (null-terminated or padded).

    This class is designed for handling strings that are padded to a
    fixed length, typically with zero-padding or any custom padding
    character. It is useful for encoding and decoding data that follows
    the C-style string conventions, where strings are often represented
    by a fixed length, and padding (e.g., null bytes) is used to fill
    the remaining space.

    Example usage:

    >>> cstring = CString(10, encoding='utf-8') # encoding is optional
    >>> pack(cstring, "Hello, World!")
    b"Hello, World\\x00"
    >>> unpack(cstring, b"Hello, World\\x00")
    'Hello, World!'

    This class also supports direct getitem access to create a list:

    >>> cstrings = CString[10] # array of ten strings (variable size)

    :param length: The fixed length of the C-string or a callable that determines
                   the length based on the context. If `None` or `Ellipsis`, the
                   string will be unpacked until the padding character is found.
    :param encoding: The encoding to use for converting the byte data to a string.
                    Defaults to `utf-8`. If `None`, the default encoding is used.
    :param pad: The padding character to use (usually `0` for null padding). If a
                string is provided, it must be a single character. If not specified,
                defaults to `0` (null byte).
    """

    __slots__ = ("encoding", "pad", "_raw_pad")

    def __init__(
        self,
        length=None,
        encoding=None,
        pad=None,
    ) -> None:
        """
        Initialize the String field with a fixed length or a length determined by a context lambda.

        :param length: The fixed length or a context lambda to determine the length dynamically.
        :param encoding: The encoding to use for string encoding/decoding (default is UTF-8).
        """
        self.length = length or ...
        self.encoding = encoding or "utf-8"
        self.pad = pad or 0
        if isinstance(self.pad, str):
            if len(self.pad) != 1:
                raise ValueError(
                    f"Invalid padding {pad!r}. Padding must be a single character."
                )
            self.pad = ord(self.pad)

        if not isinstance(self.pad, int):
            raise ValueError(
                f"Invalid padding {pad!r}. Padding must be a an integer or a single character."
            )
        self._raw_pad = self.pad.to_bytes(1, byteorder="big")

    def __class_getitem__(cls, dim) -> Field:
        """
        Allows indexing for `CString` class to support field dimensioning.

        :param dim: The dimension or index of the field.
        :return: A new instance of `CString` with the specified dimension.
        """
        return CString(...)[dim]

    def __size__(self, context):
        """
        Returns the size of the `CString` field.

        :param context: The context used to determine the size.
        :return: The length of the field in bytes.
        """
        return self.length(context) if callable(self.length) else self.length

    def __type__(self) -> type:
        """
        Returns the type associated with the `CString` field.

        :return: The `str` type, as the field stores string data.
        """
        return str

    def pack_single(self, obj: str, context) -> None:
        """
        Pack a single string into the stream with padding.

        This method encodes the string into bytes and pads it to the
        fixed length, if applicable. If the length is dynamic or
        unspecified, padding is added until the total length is achieved.

        :param obj: The string to pack into the stream.
        :param context: The current context.
        :raises ValidationError: If the string is too long for the fixed length.
        """
        encoded = obj.encode(self.encoding)
        stream = context[CTX_STREAM]
        if self.length is not Ellipsis:
            length = self.__size__(context)
            obj_length = len(obj)
            if obj_length > length:
                raise ValidationError(
                    f"String {obj!r} is too long for the fixed length of {length} bytes."
                )
            stream.write(encoded)
            stream.write(self._raw_pad * (length - obj_length))
        else:
            stream.write(encoded)
            stream.write(self._raw_pad)

    def unpack_single(self, context):
        """
        Unpack a single C-style string from the stream.

        This method reads bytes from the stream until a padding
        character is found. If the length is fixed, it will unpack
        exactly that number of bytes. If length is unspecified, it
        unpacks until the padding character is encountered.

        :param context: The current context, which provides access to the
                        stream and other necessary metadata.
        :return: The unpacked string, stripped of padding, decoded with
                 the specified encoding.
        """
        if self.length is Ellipsis:
            # Parse actual C-String
            stream = context[CTX_STREAM]
            data = bytearray()
            while True:
                value = stream.read(1)
                if not value or value[0] == self._raw_pad[0]:
                    break
                data.extend(value)
            value = bytes(data)
        else:
            length = self.__size__(context)
            value: bytes = context[CTX_STREAM].read(length)

        return value.rstrip(self._raw_pad).decode(self.encoding)


class ConstString(Const):
    """
    A specialized constant field for handling fixed string values.

    This class is used to define a field that always holds a constant
    string value. The value is validated during both encoding and
    decoding to ensure that the expected constant string is preserved.
    The string can optionally be encoded in a specific encoding.

    Additionally, this class provides a type converter that can be used
    to convert a string annotation to a `ConstString` field.

    >>> @struct
    ... class MyStruct:
    ...     value: "const str" # <-- will be converted to ConstString

    Note that the type declaration above won't work if you enable
    annotation evaluation.

    Examples:

    >>> # Define a constant string value to enforce during encoding/decoding
    >>> const_str = ConstString("Hello, World!")
    >>> pack(None, const_str, as_field=True)
    b"Hello, World!"
    >>> unpack(const_str, b"Hello, World!", as_field=True)
    "Hello, World!"

    :param value: The constant string value to be encoded/decoded.
    :param encoding: The encoding to use for the string (default is UTF-8).
                     If `None`, the default system encoding is used.
    """

    __slots__ = ()

    def __init__(self, value: str, encoding=None) -> None:
        if not isinstance(value, str):
            raise TypeError("value must be a string")

        struct = String(len(value), encoding)
        super().__init__(value, struct)
        # REVISIT: rework bitfield
        self.__bits__ = len(value) * 8


@registry.TypeConverter(str)
def _str_type_converter(annotation: str, kwargs: dict) -> ConstString:
    return ConstString(annotation)


registry.annotation_registry.append(_str_type_converter)


class ConstBytes(Const):
    """
    A constant field for handling fixed bytes values.

    This class is used to define a field that always holds a constant
    bytes value. The value is validated during both encoding and
    decoding to ensure that the expected constant byte sequence is
    preserved.

    This class provides a type converter that can be used to convert
    a bytes annotation to a `ConstBytes` field.

    >>> @struct
    ... class MyStruct:
    ...     value: b"const bytes" # <-- will be converted to ConstBytes

    :param value: The constant bytes value to be encoded/decoded.
    """

    __slots__ = ()

    def __init__(self, value: bytes) -> None:
        super().__init__(value, Bytes(len(value)))
        self.__bits__ = len(value) * 8


@registry.TypeConverter(bytes)
def _bytes_type_converter(annotation: bytes, kwargs: dict) -> ConstBytes:
    return ConstBytes(annotation)


registry.annotation_registry.append(_bytes_type_converter)


class Computed(FieldStruct):
    """
    A specialized field for representing computed or dynamically calculated values.

    This class is used to define a field that holds a computed value. The value
    can either be a constant or a function (lambda) that computes the value
    dynamically based on the context during packing or unpacking.

    The computed value is not directly stored; instead, it is evaluated using
    the provided lambda function or constant value during encoding and decoding.

    Examples:

    >>> # Define a computed field with a constant value
    >>> computed = Computed(42)
    >>> unpack(computed, b"\\x00\\x01", as_field=True)
    42
    >>> pack(None, computed, as_field=True)
    b"" # won't affect the stream

    :param value: A constant value or a lambda function that computes the value
                  based on the context.
    """

    __slots__ = ("value",)

    def __init__(self, value) -> None:
        self.value = value
        self.__bits__ = 0

    def __type__(self):
        """
        Return the type of the computed field.

        :return: `Any` if callable, otherwise the type of the constant value.
        """
        return Any if callable(self.value) else type(self.value)

    def __pack__(self, obj, context) -> None:
        """
        No packing is needed for computed fields, as the value is computed dynamically.

        This function exists to satisfy the `FieldStruct` interface, but does not
        perform any packing.

        :param obj: The object to pack.
        :param context: The current context.
        """
        pass

    def __size__(self, context) -> int:
        """
        Return the size of the computed field.

        Since the field is computed dynamically, the size is fixed as `0`.

        :param context: The current context.
        :return: Always returns `0` for computed fields.
        """
        return 0

    def __unpack__(self, context):
        """
        Unpack the computed value based on the context.

        If the value is a callable (lambda), it will be invoked with the context
        to calculate the value. Otherwise, the constant value is returned directly.

        :param context: The current context.
        :return: The computed value or the constant value.
        """
        return self.value(context) if callable(self.value) else self.value

    def pack_single(self, obj, context) -> None:
        """
        No packing is needed for computed fields.

        :param obj: The object to pack.
        :param context: The current context.
        """
        # No need for an implementation
        pass

    def unpack_single(self, context) -> None:
        """
        No unpacking is needed for computed fields.

        :param context: The current context.
        """
        # No need for an implementation
        pass


@singleton
class Pass(FieldStruct):
    """
    A specialized field that does nothing during packing and unpacking.

    This class is used to define a field that effectively "passes" through
    the packing and unpacking process without any modifications or
    transformations. It serves as a placeholder that doesn't store or process
    any data. This can be useful in cases where you need a field in a structure
    but don't want it to affect the overall packing or unpacking logic.

    The `Pass` field always returns `None` during unpacking and does not
    perform any operations during packing.

    Example usage:

    >>> @struct
    ... class MyStruct:
    ...     abc: Pass
    ...
    >>> obj = unpack(MyStruct, b"\\x00\\x01")
    MyStruct(abc=None)
    >>> pack(MyStruct(abc=None))
    b""
    """

    __slots__ = ()

    def __bits__(self) -> int:
        """
        Return the size in bits for the Pass field.

        Since the `Pass` field doesn't hold any data, it is considered to have a
        size of `0` bits.

        :return: Always returns `0` for the Pass field.
        """
        return 0

    def __type__(self) -> type:
        """
        Return the type associated with this Pass field.

        The `Pass` field doesn't have a type associated with actual data. It
        effectively represents a `None` type when unpacked.

        :return: `NoneType`.
        """
        return None.__class__

    def __pack__(self, obj, context) -> None:
        pass

    def __size__(self, context) -> int:
        return 0

    def __unpack__(self, context):
        return None

    def pack_single(self, obj, context) -> None:
        # No need for an implementation
        pass

    def unpack_single(self, context) -> None:
        # No need for an implementation
        pass


class Prefixed(FieldStruct):
    """
    A specialized field for handling data with a prefix (typically indicating length or size).

    This class is used when you want to prefix a data structure with a size or other identifying
    information. The prefix is typically packed first, followed by the actual data. The field
    allows for handling both prefixed data with a structure and simple byte sequences, depending
    on the presence of the `struct` argument. However, the returned prefix value must be
    an integer.

    Example usage:

    >>> prefixed = Prefixed(uint32, Bytes(4))
    >>> pack(b"abcd", prefixed, as_field=True)
    b"\\x00\\x00\\x00\\x04abcd"
    >>> unpack(prefixed, b"\\x00\\x00\\x00\\x04abcd", as_field=True)
    b"abcd"

    The prefix is packed and unpacked independently and is used to determine the size of the
    following data.

    **Please note that this class will have a huge impact on the resulting packing and
    unpacking time and will increase it significantly.**

    Starting from v2.4.0, this class uses the `struct` as its second parameter, moving the
    'encoding' parameter to the third parameter.

    :param prefix: The struct that defines the prefix (e.g., a length field).
    :param struct: The struct that defines the data to follow the prefix (e.g., `Bytes` or another field).
    :param encoding: The encoding to use for the prefix.
    """

    __slots__ = ("prefix", "struct", "encoding")

    def __init__(
        self,
        prefix,
        struct=None,
        encoding=None,
    ):
        self.prefix = prefix
        self.struct = struct
        self.encoding = encoding
        # Support str as second argument
        if isinstance(struct, str):
            warnings.warn(
                "Using a string as the second argument is deprecated since 2.4.0. Please use encoding=... instead!",
                DeprecationWarning,
            )
            self.encoding, self.struct = struct, None

    def __type__(self):
        """
        Return the type associated with this Prefixed field.

        If the `struct` is provided, it returns the type of the `struct`. Otherwise, it returns
        `bytes` as the default type for a prefixed field.

        :return: The type associated with the Prefixed field (`bytes` or the type of the struct).
        """
        return bytes if self.struct is None else self.struct.__type__()

    def __size__(self, context) -> int:
        """
        Prefixed fields do not have a fixed size.

        The size of a `Prefixed` field is dynamic, as it depends on the prefix and the data length,
        which can vary. Therefore, calling this method will raise an error.

        :raises DynamicSizeError: Always raises an error as the size is not fixed.
        """
        raise DynamicSizeError("Prefixed does not store a size", context)

    def pack_single(self, obj, context) -> None:
        """
        Pack a single object into the stream, with the prefix indicating the size.

        This method first packs the length of the data (the prefix), and then writes the actual data.

        :param obj: The object to pack (should be a byte sequence).
        :param context: The current context.
        """
        if self.struct is not None:
            data = BytesIO()
            with WithoutContextVar(context, CTX_STREAM, data):
                self.struct.__pack__(obj, context)

            context[CTX_SEQ] = False
            obj = data.getvalue()

        elif self.encoding:
            obj = obj.encode(self.encoding)

        self.prefix.__pack__(len(obj), context)
        context[CTX_STREAM].write(obj)

    def unpack_single(self, context):
        """
        Unpack a single object from the stream, using the prefix to determine the size.

        This method first unpacks the prefix (which indicates the size), then reads the data
        from the stream based on that size. If a `struct` is provided, the data is then passed
        through the specified structure for further unpacking.

        :param context: The current context.
        :return: The unpacked object, which is either raw bytes or the data structure.
        """
        size = self.prefix.__unpack__(context)
        if not isinstance(size, int):
            raise TypeError(f"Expected int, got {type(size)}")

        data = context[CTX_STREAM].read(size)
        obj = data
        if self.struct is not None:
            with (
                WithoutContextVar(context, CTX_STREAM, BytesIO(data)),
                WithoutContextVar(context, CTX_SEQ, False),
            ):
                obj = self.struct.__unpack__(context)
        elif self.encoding:
            obj = data.decode(self.encoding)
        return obj


class Int(FieldStruct):
    """Generic Integer

    This class handles packing and unpacking integer values with a specified bit size, either
    signed or unsigned. The size of the integer is defined in bits, and the class ensures the
    appropriate byte representation is used for packing and unpacking.

    :param bits: The bit width of the integer (e.g., 8, 16, 32, 64).
    :param signed: Whether the integer is signed (default is True).
    """

    __slots__ = ("signed", "size")

    def __init__(self, bits: int, signed=True) -> None:
        self.signed = signed
        self.__bits__ = bits
        if not isinstance(bits, int):
            raise ValueError(f"Invalid int size: {bits!r} - expected int")
        self.size = self.__bits__ // 8
        self.__byteorder__ = SysNative

    def __repr__(self) -> str:
        name = "int"
        if not self.signed:
            name = f"u{name}"

        return f"<{name}{self.__bits__}>"

    def __type__(self) -> type:
        """
        Return the type associated with this Int field.

        :return: The type of the field, which is `int`.
        """
        return int

    def __size__(self, context) -> int:
        """
        Return the size of the integer in bytes.

        :param context: The current context.
        :return: The size of the integer in bytes.
        """
        return self.size

    def pack_single(self, obj: int, context) -> None:
        """
        Pack a single integer value into the stream.

        This method converts the integer value into a byte representation using the specified
        byte order. The byte order is determined from the context's field order
        (either little-endian or big-endian).

        :param obj: The integer value to pack.
        :param context: The current context, which provides the byte order (little-endian or big-endian).
        :raises ValueError: If the integer is too large or small for the given bit width.
        """
        field = context.get(CTX_FIELD)
        is_little = (
            field.order.ch if field else self.__byteorder__.ch
        ) == LITTLE_ENDIAN_FMT
        context[CTX_STREAM].write(
            obj.to_bytes(
                length=self.size,
                byteorder="little" if is_little else "big",
                signed=self.signed,
            )
        )

    def unpack_single(self, context) -> int:
        """
        Unpack a single integer value from the stream.

        This method reads a byte sequence from the stream, interprets it as an integer using the
        specified byte order and signedness, and returns the integer value.

        :param context: The current context, which provides the byte order (little-endian or big-endian).
        :return: The unpacked integer value.
        :raises ValueError: If the data cannot be unpacked as an integer of the specified size.
        """
        field = context.get(CTX_FIELD)
        is_little = (
            field.order.ch if field else self.__byteorder__.ch
        ) == LITTLE_ENDIAN_FMT

        return int.from_bytes(
            context[CTX_STREAM].read(self.size),
            "little" if is_little else "big",
            signed=self.signed,
        )


class UInt(Int):
    """Generic unsigned integer."""

    __slots__ = ()

    def __init__(self, bits: int) -> None:
        super().__init__(bits, signed=False)


int24 = Int(24)
uint24 = UInt(24)


class Aligned(FieldStruct):
    """
    Alignment of a struct (before or after)

    This class ensures that the associated `struct` is properly aligned according to
    the specified alignment. It allows for padding before or after the field, ensuring
    that the field starts or ends at a memory address that is a multiple of the alignment.

    Example usage:

    >>> @struct
    ... class Format:
    ...     a: Aligned(int16, alignment=4, after=True)
    ...     b: uint8
    ...
    >>> unpack(Format, b"\\x00\\x01\\xFF\\xFF\\x01")
    Traceback (most recent call last):
    ...
    ValueError: Expected 2 bytes of padding (value=0), got 0
    >>> unpack(Format, b"\\x00\\x01\\x00\\x00\\x01")
    Format(a=256, b=1)

    This example shows a `Format` structure where:
    - `a` is an aligned integer (with 4-byte alignment) and padding is applied after the field.
    - `b` is an unsigned byte (`uint8`), which follows `a` after the padding.

    :param struct: The structure that is to be aligned.
    :param alignment: The alignment value in bytes, which must be a power of 2.
                      This can be an integer or a context lambda for dynamic alignment.
    :param after: If `True`, padding is applied after the structure (align after).
    :param before: If `True`, padding is applied before the structure (align before).
    :param filler: The byte value to use for padding. It can be an integer or a string
                   (default is zero-padding). If no filler is provided, `b"\x00"` will
                   be used as the padding byte.
    :raises ValueError: If neither `before` nor `after` is specified, if the filler is not a
                        single byte, or if the padding does not match the expected value.
    :raises DynamicSizeError: If dynamic alignment is used and the size cannot be determined.
    """

    __slots__ = ("struct", "_after", "_before", "_filler", "alignment")

    def __init__(
        self,
        struct,
        alignment,
        after=False,
        before=False,
        filler=None,
    ) -> None:
        if not before and not after:
            raise ValueError("Must specify either before or after")

        self.struct = struct
        self.alignment = alignment
        self._after = after
        self._before = before
        if filler is None:
            self._filler = 0x00
        else:
            if isinstance(filler, str):
                self._filler = ord(filler)
            else:
                self._filler = filler

        if not isinstance(self._filler, int):
            raise ValueError(f"Filler must be a single byte - got {filler!r}")

    def __type__(self):
        """
        Return the type associated with this aligned field.

        :return: The type of the struct (usually `bytes` or another custom type).
        """
        return self.struct.__type__()

    def __size__(self, context) -> int:
        """
        Calculate the size of the aligned field, accounting for padding based on the alignment.

        :param context: The current context.
        :return: The size of the aligned field, including padding if necessary.
        :raises DynamicSizeError: If the alignment is dynamic and cannot be determined.
        """
        if callable(self.alignment):
            raise DynamicSizeError(
                "Aligned fields with dynamic alignment don't have a fixed size"
            )

        struct_size = self.struct.__size__(context)
        return struct_size + (self.alignment - (struct_size % self.alignment))

    def unpack_alignment(self, context):
        """
        Unpack padding for the alignment, verifying that the correct amount of padding is present.

        :param context: The current context.
        :raises ValueError: If the padding does not match the expected value.
        """
        value = self.alignment(context) if callable(self.alignment) else self.alignment
        if not isinstance(value, int):
            raise ValueError(f"Alignment must be an integer - got {value!r}")

        if value % 2 != 0:
            raise ValueError(f"Alignment must be a power of 2 - got {value!r}")

        stream = context[CTX_STREAM]
        pos = stream.tell()
        size = value - (pos % value)
        data = stream.read(size)
        if data.count(self._filler) != size:
            raise ValueError(
                f"Expected {size} bytes of padding (value={self._filler!r}), got {data.count(self._filler)}"
            )

    def unpack_single(self, context):
        """
        Unpack a single aligned field from the stream.

        This method ensures that padding is applied before or after the field as needed
        based on the `before` and `after` parameters.

        :param context: The current context.
        :return: The unpacked structure, properly aligned.
        """
        if self._before:
            self.unpack_alignment(context)
        obj = self.struct.__unpack__(context)
        if self._after:
            self.unpack_alignment(context)
        return obj

    def pack_alignment(self, context):
        """
        Apply padding for the alignment before or after the structure, depending on
        the `before` and `after` settings.

        :param context: The current context.
        """
        value = self.alignment(context) if callable(self.alignment) else self.alignment
        stream = context[CTX_STREAM]
        size = value - (stream.tell() % value)
        stream.write(bytes([self._filler] * size))

    def pack_single(self, obj, context) -> None:
        """
        Pack a single aligned field into the stream, applying padding if necessary.

        :param obj: The structure to pack.
        :param context: The current context.
        """
        if self._before:
            self.pack_alignment(context)
        self.struct.__pack__(obj, context)
        if self._after:
            self.pack_alignment(context)


def align(alignment):
    """
    Create a context lambda to calculate the alignment padding required at the current stream position.

    This function generates a lambda that can be used to compute the amount of padding required
    to ensure that the next structure in the stream is aligned to the specified alignment.

    The alignment value can either be a fixed integer (representing the alignment in bytes)
    or a context lambda that computes the alignment dynamically based on the current context.

    Example usage:

    >>> @struct
    ... class Format:
    ...     a: uint8
    ...     b: padding[align(4)]
    ...
    >>> unpack(Format, b"\\x01\\x00\\x00\\x00")
    Format(a=1, b=None)

    :param alignment: The alignment value in bytes, which must be a power of 2.
                      This can be either an integer or a context lambda.
    :return: A context lambda function that returns the number of bytes to align the next structure.
    """

    def _get_aligned_size(context: _ContextLike):
        pos = context[CTX_STREAM].tell()
        value = alignment(context) if callable(alignment) else alignment
        return value - (pos % value)

    return _get_aligned_size


class Lazy(FieldStruct):
    """
    A lazy field struct that defers the creation of the underlying struct until it is needed.

    This class allows the definition of a field where the underlying struct is not created until
    the field is actually accessed. This can help with optimization in scenarios where certain
    fields may not always be used.

    The underlying struct is defined by a callable that returns an instance of a struct.
    The `struct_fn` function is invoked to generate the struct when the field is accessed.

    Example usage:

    >>> @struct
    ... class Format:
    ...     a: Lazy(lambda: SecondFormat)  # Lazy instantiation of the struct
    ...     b: uint8
    ...
    >>> @struct
    ... class SecondFormat:
    ...     a: uint8
    ...
    >>> unpack(Format, b"\\x01\\x02\\xFF")
    Format(a=SecondFormat(a=1), b=255)

    :param struct: A callable that returns the underlying struct, which will be lazily created
                   when the field is accessed.
    """

    def __init__(self, struct) -> None:
        if not callable(struct):
            raise TypeError(f"struct must be a callable - got {struct!r}")

        self.struct_fn = struct

    @cached_property
    def struct(self) -> _StructLike:
        """
        Get the underlying struct by invoking the callable.

        This method ensures that the struct is only created once, and it is cached for subsequent use.

        :return: The underlying struct.
        :rtype: _StructLike
        """
        return self.struct_fn()

    def __bits__(self):
        """
        Get the bit representation of the Lazy struct by delegating to the underlying struct.

        :return: The bit representation of the struct.
        :rtype: str
        """
        return self.struct.__bits__()

    def __type__(self):
        """
        Get the type associated with the Lazy struct by delegating to the underlying struct.

        :return: The type associated with the struct.
        :rtype: type
        """
        return self.struct.__type__()

    def __size__(self, context) -> int:
        """
        Get the size of the Lazy struct by delegating to the underlying struct.

        :param context: The context for size calculation.
        :return: The size of the struct.
        :rtype: int
        """
        return self.struct.__size__(context)

    def pack_single(self, obj, context) -> None:
        """
        Pack a single value using the Lazy struct by delegating to the underlying struct.

        :param obj: The value to pack.
        :param context: The context for packing.
        """
        self.struct.__pack__(obj, context)

    def unpack_single(self, context):
        """
        Unpack a single value using the Lazy struct by delegating to the underlying struct.

        :param context: The context for unpacking.
        :return: The unpacked value.
        """
        return self.struct.__unpack__(context)


@singleton
class Uuid(FieldStruct):
    """
    A field for handling UUID values.

    This class is responsible for serializing and deserializing UUIDs to and
    from their binary representation. The UUID is represented as a 16-byte
    field (128 bits) and supports both little-endian and big-endian formats.

    Example usage:

    >>> unpack(Uuid, b'e3215476e89b12d3', as_field=True)
    UUID('65333231-3534-3736-6538-396231326433')

    :param value: The UUID value, which is typically a `UUID` object from the
                  `uuid` module.
    :param context: The context for packing and unpacking the UUID. The
                    context includes stream order information for byte
                    order (little-endian or big-endian).
    """

    __slots__ = ()

    def __type__(self) -> type:
        """
        Get the type associated with the UUID field.

        :return: The type `UUID`.
        :rtype: type
        """
        return UUID

    def __size__(self, context) -> int:
        """
        Get the size of the UUID field.

        :param context: The context for size calculation.
        :return: The size of the UUID in bytes (16 bytes).
        :rtype: int
        """
        return 16

    def __bits__(self) -> int:
        """
        Get the bit representation of the UUID field.

        :return: The bit size of the UUID field (128 bits).
        :rtype: int
        """
        return 128

    def __pack__(self, obj: UUID, context) -> None:
        """
        Pack a UUID object into the stream.

        Depending on the byte order (little-endian or big-endian) specified
        in the context, the UUID will be packed in the corresponding byte order.

        :param obj: The `UUID` object to pack.
        :param context: The current context, which includes the stream and byte order.
        """
        is_le = context[CTX_FIELD].order is LittleEndian
        super().__pack__(obj.bytes_le if is_le else obj.bytes, context)

    def __unpack__(self, context) -> UUID:
        """
        Unpack a UUID from the stream.

        Depending on the byte order (little-endian or big-endian) specified
        in the context, the UUID will be unpacked in the corresponding byte order.

        :param context: The current context, which includes the stream and byte order.
        :return: The unpacked `UUID` object.
        :rtype: UUID
        """
        is_le = context[CTX_FIELD].order is LittleEndian
        data = context[CTX_STREAM].read(16)
        return UUID(bytes_le=data) if is_le else UUID(bytes=data)


class AsLengthRef:
    """
    A special field for automatically determining the length of a field
    based on a target field.

    This field is used to pack and unpack a length value for a target field.
    When packing, the length of the target field is determined by getting the
    length of the target field's value from the context. When unpacking, the
    unpacked value is used as the length of the target field.

    Example usage:

    >>> @struct
    ... class MyStruct:
    ...     length: AsLengthRef("length", "data", uint16) = 0
    ...     a: uint8
    ...     data: Bytes(this.length)
    ...
    >>> pack(MyStruct(data=b"Hello, world!", a=1))
    b'\\x00\\x0c\\x01Hello, world!'

    :param target: The target field name to determine the length from.
    :param struct: The struct definition of the length field.
    """

    __slots__ = (
        "struct",
        "target",
        "name",
    )

    def __init__(self, name: str, target: str, struct=None) -> None:
        self.struct = struct
        self.name = f"_obj.{name}"
        self.target = f"_obj.{target}"

    def __mod__(self, other) -> Self:
        self.struct = other
        return self

    def __rmod__(self, other) -> Self:
        self.struct = other
        return self

    def __type__(self) -> type:
        return int

    def __size__(self, context) -> int:
        if self.struct is None:
            return 0
        return self.struct.__size__(context)

    def __bits__(self) -> int:
        if self.struct is None:
            return 0
        return self.struct.__bits__()

    def __pack__(self, obj, context):
        # object is optional
        if self.struct is None:
            raise ValueError("struct is not defined")

        target_obj = context.__context_getattr__(self.target)
        length = len(target_obj)
        context.__context_setattr__(self.name, length)
        self.struct.__pack__(length, context)

    def __unpack__(self, context):
        if self.struct is None:
            raise ValueError("struct is not defined")

        length = self.struct.__unpack__(context)
        context.__context_setattr__(self.name, length)
        return length

    def __repr__(self) -> str:
        name = self.target.removeprefix("_obj.")
        return f"<LengthRef of .{name}>"



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from enum import Enum

from caterpillar.abc import _EndianLike, _ArchLike, _SupportsSetEndian, _OT

class ByteOrder(_EndianLike):
    name: str
    ch: str

    class Alignment(Enum):
        NONE = 0
        NATIVE = 1

    alignment: Alignment

    class Size(Enum):
        STANDARD = 0
        NATIVE = 1

    size: Size

    def apply(self, other: ByteOrder) -> None: ...
    def __add__(self, other: _SupportsSetEndian[_OT]) -> _OT: ...
    def __or__(self, other: ByteOrder): ...
    def __init__(
        self,
        name: str,
        ch: str,
        alignment: Alignment | None = ...,
        size: Size | None = ...,
    ) -> None: ...

LITTLE_ENDIAN_FMT: str
Native: ByteOrder = ...
BigEndian: ByteOrder = ...
LittleEndian: ByteOrder = ...
NetEndian: ByteOrder = ...
SysNative: ByteOrder = ...

def byteorder(obj, default: _EndianLike | None = None) -> _EndianLike: ...

class Arch(_ArchLike):
    name: str
    ptr_size: int

    def __init__(self, name: str, ptr_size: int) -> None: ...

system_arch: Arch = ...
x86: Arch = ...
x86_64: Arch = ...
ARM: Arch = ...
ARM64: Arch = ...
AARCH64: Arch = ...
PowerPC: Arch = ...
PowerPC64: Arch = ...
MIPS: Arch = ...
MIPS64: Arch = ...
SPARC: Arch = ...
SPARC64: Arch = ...
RISC_V64: Arch = ...
RISC_V: Arch = ...
AMD: Arch = ...
AMD64: Arch = ...



from hachoir.field import *
from hachoir.stream import StringInputStream
from hachoir.core.bits import LITTLE_ENDIAN


class Entry(FieldSet):
    endian = LITTLE_ENDIAN

    def createFields(self):
        yield UInt8(self, "num1")
        yield UInt24(self, "num2")
        yield Bit(self, "bool1")
        yield Bits(self, "num4", 3)
        yield PaddingBits(self, "_", 4)
        for _ in range(3):
            yield UInt8(self, "fixedarray[]")
        yield CString(self, "name1")
        yield PascalString8(self, "name2")


class Format(Parser):
    endian = LITTLE_ENDIAN

    def createFields(self):
        yield UInt32(self, "count")
        for _ in range(self["count"].value):
            yield Entry(self, "entry[]")


if __name__ == "__main__":
    import sys
    import timeit

    try:
        from rich import print
    except ImportError:
        pass

    with open(sys.argv[1], "rb") as fp:
        data = fp.read()

    time = (
        timeit.timeit(lambda: list(Format(StringInputStream(data))), number=1000) / 1000
    )
    print("[bold]Parsing measurements:[/]")
    print(f"[bold]default[/]  {time:.10f} sec/call")



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from io import RawIOBase

from caterpillar.context import CTX_STREAM


class IOHook(RawIOBase):
    """
    A custom I/O stream wrapper that allows hooks to be installed for various
    stages of stream interaction, such as initialization, reading and writing.

    This class can be used to augment the functionality of an existing stream (e.g.,
    file streams, memory buffers, or network sockets) by attaching custom logic for
    handling data as it is read or written.

    :param io: The underlying I/O stream to which the hooks are applied.
    :type io: RawIOBase
    :param init: Optional hook function to be called during initialization.
    :type init: Optional[HookInit]
    :param update: Optional hook function to modify data before or after reading/writing.
    :type update: Optional[HookUpdate]
    :param read: Optional hook function to be applied to data after reading.
    :type read: Optional[HookRead]
    :param write: Optional hook function to be applied to data before writing.
    :type write: Optional[HookWrite]
    :param finish: Optional hook function to be called when the stream is finished.
    :type finish: Optional[HookFinish]
    """

    def __init__(
        self, io, init=None, update=None, read=None, write=None, finish=None
    ) -> None:
        # NOTE: no validation here if _io is valid, because
        # self.init will set it
        self._io = io
        self._init = init
        self._update = update
        self._read = read
        self._write = write
        self._finish = finish
        self._context = None

    def assert_context_set(self) -> None:
        """
        Ensures that the context has been set before any I/O operations are performed.

        :raises ValueError: If the context is not set.
        """
        if self._context is None:
            raise ValueError("Context is not set")

    def init(self, context) -> None:
        """
        Initialize the I/O hook with the provided context. This triggers the
        `init` hook, if available, and sets up the context for subsequent operations.

        :param context: The context to be used for initializing the stream.
        :type context: _ContextLike
        """
        if self._init:
            self._init(context)
        self._context = context
        self._io = context[CTX_STREAM]
        self._context[CTX_STREAM] = self

    def finish(self, context) -> None:
        """
        Finalize the I/O hook by calling the `finish` hook (if provided) and
        restoring the original I/O stream in the context.

        :param context: The context used during the finalization.
        :type context: _ContextLike
        """
        if self._finish:
            self._finish(context)

        context[CTX_STREAM] = self._io
        self._context = None

    def seekable(self) -> bool:
        """
        Checks if the underlying I/O stream is seekable.

        :return: True if the stream supports seeking, otherwise False.
        :rtype: bool
        """
        return self._io.seekable()

    def readable(self) -> bool:
        """
        Checks if the underlying I/O stream is readable.

        :return: True if the stream supports reading, otherwise False.
        :rtype: bool
        """
        return super().readable()

    def read(self, size: int = -1):
        """
        Read data from the stream, applying the optional hooks (if any).

        - The `update` hook, if provided, will be called to modify the data.
        - The `read` hook, if provided, will be called to further modify the data.

        :param size: The number of bytes to read, defaults to -1 (read until EOF).
        :type size: int
        :return: The read data, possibly modified by the hooks.
        :rtype: Union[bytes, None]
        """
        self.assert_context_set()
        data = self._io.read(size)
        if data is None:
            return

        if self._update:
            data = self._update(data, self._context) or data

        if self._read:
            data = self._read(data, self._context) or data

        return data

    def write(self, b, /):
        """
        Write data to the stream, applying the optional hooks (if any).

        - The `update` hook, if provided, will be called to modify the data before writing.
        - The `write` hook, if provided, will be called to modify the data before the final write operation.

        :param b: The data to write.
        :type b: bytes
        :return: The number of bytes written to the underlying stream.
        :rtype: Union[int, None]
        """
        self.assert_context_set()
        if self._update:
            b = self._update(b, self._context) or b

        if self._write:
            b = self._write(b, self._context) or b

        return self._io.write(b)

    def writable(self) -> bool:
        """
        Checks if the underlying I/O stream is writable.

        :return: True if the stream supports writing, otherwise False.
        :rtype: bool
        """
        return self._io.writable()

    def tell(self) -> int:
        """
        Returns the current position of the underlying stream.

        :return: The current position in the stream.
        :rtype: int
        """
        return self._io.tell()

    def seek(self, offset: int, whence: int = 0) -> int:
        """
        Move the cursor to a new position in the underlying stream.

        :param offset: The offset to move.
        :param whence: The reference point for the offset (default is 0, SEEK_SET).
        :return: The new position in the stream.
        :rtype: int
        """
        return self._io.seek(offset, whence)



# type: ignore
from caterpillar.py import struct, BigEndian, this, unpack, uint8

try:
    from rich import print
except ImportError:
    pass


@struct(order=BigEndian)
class Format:
    magic: b"BMP"
    width: uint8
    height: uint8
    pixels: uint8[this.width * this.height]


@struct(order=BigEndian)
class Format2:
    magic: b"BMP"
    width: uint8
    height: uint8
    pixels: uint8[...]


print(unpack(Format[...], b"BMP\x02\x02\x00\x01\x02\x03BMP\x00\x00"))
# Prints:
# [Format(magic=b'BMP', width=2, height=2, pixels=[0, 1, 2, 3]), Format(magic=b'BMP', width=0, height=0, pixels=[])]
# and
print(unpack(Format2[...], b"BMP\x02\x02\x00\x01\x02\x03BMP\x00\x00"))
# prints:
# [Format2(magic=b'BMP', width=2, height=2, pixels=[0, 1, 2, 3, 66, 77, 80, 0, 0])]



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from types import FrameType
from typing import (
    Any,
    Callable,
    Final,
    Protocol,
    Union,
    dataclass_transform,
    type_check_only,
)
from typing_extensions import Self, override
from caterpillar.abc import _ContextLike, _ContextLambda
from caterpillar.options import Flag

CTX_PARENT: str
CTX_OBJECT: str
CTX_OFFSETS: str
CTX_STREAM: str
CTX_FIELD: str
CTX_VALUE: str
CTX_POS: str
CTX_INDEX: str
CTX_PATH: str
CTX_SEQ: str
CTX_ARCH: str
CTX_ROOT: str

@type_check_only
class ContextFactory(Protocol):
    def __call__(self, **kwds) -> _ContextLike: ...

O_CONTEXT_FACTORY: Flag[type[_ContextLike] | ContextFactory]

class Context(dict[str, Any]):
    @override
    def __setattr__(self, key: str, value: Any) -> None: ...
    @override
    def __getattribute__(self, key: str) -> Any: ...
    def __context_getattr__(self, path: str) -> Any: ...
    def __context_setattr__(self, path: str, value: Any) -> None: ...

class ExprMixin:
    def __add__(self, other: Any) -> BinaryExpression: ...
    def __sub__(self, other: Any) -> BinaryExpression: ...
    def __mul__(self, other: Any) -> BinaryExpression: ...
    def __floordiv__(self, other: Any) -> BinaryExpression: ...
    def __truediv__(self, other: Any) -> BinaryExpression: ...
    def __mod__(self, other: Any) -> BinaryExpression: ...
    def __pow__(self, other: Any) -> BinaryExpression: ...
    def __xor__(self, other: Any) -> BinaryExpression: ...
    def __and__(self, other: Any) -> BinaryExpression: ...
    def __or__(self, other: Any) -> BinaryExpression: ...
    def __rshift__(self, other: Any) -> BinaryExpression: ...
    def __lshift__(self, other: Any) -> BinaryExpression: ...
    __div__ = __truediv__
    def __radd__(self, other: Any) -> BinaryExpression: ...
    def __rsub__(self, other: Any) -> BinaryExpression: ...
    def __rmul__(self, other: Any) -> BinaryExpression: ...
    def __rfloordiv__(self, other: Any) -> BinaryExpression: ...
    def __rtruediv__(self, other: Any) -> BinaryExpression: ...
    def __rmod__(self, other: Any) -> BinaryExpression: ...
    def __rpow__(self, other: Any) -> BinaryExpression: ...
    def __rxor__(self, other: Any) -> BinaryExpression: ...
    def __rand__(self, other: Any) -> BinaryExpression: ...
    def __ror__(self, other: Any) -> BinaryExpression: ...
    def __rrshift__(self, other: Any) -> BinaryExpression: ...
    def __rlshift__(self, other: Any) -> BinaryExpression: ...
    def __neg__(self) -> UnaryExpression: ...
    def __pos__(self) -> UnaryExpression: ...
    def __invert__(self) -> UnaryExpression: ...
    def __contains__(self, other: Any) -> BinaryExpression: ...
    def __gt__(self, other: Any) -> BinaryExpression: ...
    def __ge__(self, other: Any) -> BinaryExpression: ...
    def __lt__(self, other: Any) -> BinaryExpression: ...
    def __le__(self, other: Any) -> BinaryExpression: ...
    def __eq__(self, other: Any) -> BinaryExpression: ...
    def __ne__(self, other: Any) -> BinaryExpression: ...

class ConditionContext:
    func: Union[_ContextLambda[bool], bool]
    annotations: dict
    namelist: list[str]
    depth: int
    def __init__(
        self, condition: _ContextLambda[bool] | bool, depth: int = 2
    ) -> None: ...
    def getframe(self, num: int, msg: str | None = None) -> FrameType: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, *_) -> None: ...

@dataclass_transform()
class BinaryExpression(ExprMixin, _ContextLambda):
    operand: Callable[[Any, Any], Any]
    left: Any | _ContextLambda
    right: Any | _ContextLambda
    def __call__(self, context: _ContextLike) -> Any: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, *_) -> None: ...

@dataclass_transform()
class UnaryExpression(_ContextLambda):
    name: str
    operand: Callable[[Any], Any]
    value: Any | _ContextLambda
    def __call__(self, context: _ContextLike): ...
    def __enter__(self): ...
    def __exit__(self, *_) -> None: ...

class ContextPath(ExprMixin, _ContextLambda):
    path: str
    call_kwargs: dict[str, Any]
    getitem_args: list[Any]
    def __init__(self, path: str | None = None) -> None: ...
    def __call__(self, context: _ContextLike | None = None, **kwds): ...
    def __getitem__(self, key) -> Self: ...
    def __type__(self) -> type[Any]: ...
    def __getattribute__(self, key: str) -> ContextPath: ...
    @property
    def parent(self) -> ContextPath: ...

class ContextLength(ExprMixin, _ContextLambda):
    path: str
    def __init__(self, path: ContextPath) -> None: ...
    def __call__(self, context: _ContextLike | None = None, **kwds) -> Any: ...

this: Final[ContextPath]
ctx: Final[ContextPath]
parent: Final[ContextPath]
root: Final[ContextPath]



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import ipaddress

from re import Pattern

from ._base import singleton as singleton
from .common import (
    Bytes as Bytes,
    Transformer as Transformer,
    UInt as UInt,
    uint32 as uint32,
)
from caterpillar.abc import _ContextLike

IPv4Address: Transformer[ipaddress.IPv4Address, int, ipaddress.IPv4Address, int]
IPv6Address: Transformer[ipaddress.IPv6Address, int, ipaddress.IPv6Address, int]

class MACAddress(Transformer[str | bytes, bytes, bytes, bytes]):
    DELIMITERS: Pattern
    sep: str
    def __init__(self, sep: str | None = None) -> None: ...
    def encode(self, obj: str | bytes, context: _ContextLike) -> bytes: ...
    def decode(self, parsed: bytes, context: _ContextLike) -> bytes: ...

MAC: MACAddress



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from ._mixin import FieldStruct
from caterpillar.abc import _ContextLambda, _ContextLike, _StructLike
from caterpillar.options import Flag
from typing import Dict, Generic, Optional, Type, TypeVar

PTR_STRICT: Flag[None]

_PtrValueT = TypeVar("_PtrValueT")

class pointer(Generic[_PtrValueT], int):
    obj: Optional[_PtrValueT]
    def get(self) -> _PtrValueT | None: ...

class Pointer(FieldStruct[int, pointer[_PtrValueT]]):
    model: _StructLike[_PtrValueT, _PtrValueT] | None
    struct: _StructLike[int, int] | _ContextLambda[_StructLike[int, int]]
    def __init__(
        self,
        struct: _StructLike[int, int] | _ContextLambda[_StructLike[int, int]],
        model: _StructLike[_PtrValueT, _PtrValueT] | None = None,
    ) -> None: ...
    def __mul__(
        self, model: _StructLike[_PtrValueT, _PtrValueT]
    ) -> Pointer[_PtrValueT]: ...
    def __type__(self) -> Type[pointer[_PtrValueT]]: ...
    def __size__(self, context: _ContextLike) -> int: ...
    def unpack_single(self, context: _ContextLike) -> pointer[_PtrValueT]: ...
    def pack_single(self, obj: int, context: _ContextLike) -> None: ...

UNSIGNED_POINTER_TYS: Dict
SIGNED_POINTER_TYS: Dict

def uintptr_fn(context: _ContextLike) -> _StructLike[int, int]: ...
def intptr_fn(context: _ContextLike) -> _StructLike[int, int]: ...

uintptr: Pointer[None]
intptr: Pointer[None]

class relative_pointer(pointer[_PtrValueT]):
    base: int
    @property
    def absolute(self) -> int: ...

class RelativePointer(Pointer[_PtrValueT]):
    def __type__(self) -> relative_pointer[_PtrValueT]: ...

offintptr: RelativePointer[None]
offuintptr: RelativePointer[None]



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from types import NoneType
from typing import Any, Callable, Collection, Generic, Iterable, TypeVar

from caterpillar.abc import _OptionLike

_VT = TypeVar("_VT", default=NoneType)

class Flag(Generic[_VT]):
    name: str
    value: _VT = ...
    _hash_: int
    def __init__(self, name: str, value: _VT = ...) -> None: ...
    def __hash__(self) -> int: ...

GLOBAL_STRUCT_OPTIONS: set[_OptionLike] = ...
GLOBAL_UNION_OPTIONS: set[_OptionLike] = ...
GLOBAL_FIELD_FLAGS: set[_OptionLike] = ...
GLOBAL_BITFIELD_FLAGS: set[_OptionLike] = ...

def configure(base: set[_OptionLike], *flags: _OptionLike) -> None: ...
def set_struct_flags(*flags: _OptionLike, with_union: bool = False) -> None: ...
def set_field_flags(*flags: _OptionLike) -> None: ...
def set_union_flags(*flags: _OptionLike) -> None: ...
def get_flags(obj: Any, attr: str | None = None) -> set[_OptionLike] | None: ...
def has_flag(flag: str | _OptionLike, obj: Any, attr: str | None = None) -> bool: ...
def get_flag(name: str, obj: Any, attr: str | None = None) -> _OptionLike | None: ...

S_DISCARD_UNNAMED: Flag
S_DISCARD_CONST: Flag
S_UNION: Flag
S_REPLACE_TYPES: Flag
S_EVAL_ANNOTATIONS: Flag
S_ADD_BYTES: Flag
S_SLOTS: Flag
F_KEEP_POSITION: Flag
F_DYNAMIC: Flag
F_SEQUENTIAL: Flag
F_OFFSET_OVERRIDE: Flag
O_ARRAY_FACTORY: Flag[Callable[[Iterable], Collection] | None]
B_OVERWRITE_ALIGNMENT: Flag
B_GROUP_END: Flag
B_GROUP_NEW: Flag
B_GROUP_KEEP: Flag
B_NO_AUTO_BOOL: Flag



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from io import BytesIO
from collections.abc import Iterable
from functools import partial

from caterpillar.byteorder import ByteOrder, byteorder
from caterpillar.options import Flag
from caterpillar.context import CTX_SEQ, CTX_STREAM
from caterpillar._common import unpack_seq, pack_seq, WithoutContextVar
from caterpillar.shared import getstruct

from ._base import Field


class FieldMixin:
    """A simple mixin to support operators used to create :class:`Field` instances."""

    def __or__(self, flag: Flag):
        """Creates a field *with* the given flag."""
        return Field(self, byteorder(self)) | flag

    def __xor__(self, flag: Flag):
        """Creates a field *without* the given flag."""
        return Field(self, byteorder(self)) ^ flag

    def __matmul__(self, offset):
        """Creates a field that should start at the given offset."""
        return Field(self, byteorder(self)) @ offset

    def __getitem__(self, dim):
        """Returns a sequenced field."""
        return Field(self, byteorder(self))[dim]

    def __rshift__(self, switch):
        """Inserts switch options into the new field"""
        return Field(self, byteorder(self)) >> switch

    def __floordiv__(self, condition):
        """Returns a field with the given condition"""
        return Field(self, byteorder(self)) // condition

    def __set_byteorder__(self, order: ByteOrder):
        """Returns a field with the given byteorder"""
        return Field(self, order=order)

    def __rsub__(self, bits):
        """Returns a field with the given bit count"""
        return Field(self, byteorder(self), bits=bits)

    def __and__(self, other):
        """Returns a chain with the next element added at the end"""
        if isinstance(other, Chain):
            return other & self
        return Chain(self, other)


class FieldStruct(FieldMixin):
    """
    A mix-in class combining the behavior of _StructLike with additional
    functionality for packing and unpacking structured data.
    """

    __slots__ = {
        "__byteorder__": (
            """
            An internal field used to measure the byte order of this struct.

            Note that this field will be used during processing only and not during
            parsing or building data. In addition, the actual byte order should be
            retrieved using the :class:`Field` instance within the context.
            """
        ),
        "__bits__": "TBD",
    }

    def pack_single(self, obj, context) -> None:
        """
        Abstract method to pack a single element.

        :param obj: The element to pack.
        :type obj: Any
        :param context: The current operation context.
        :type context: _ContextLike
        :raises NotImplementedError: This method must be implemented by subclasses.
        """
        raise NotImplementedError

    def unpack_single(self, context):
        """
        Abstract method to unpack a single element.

        :param context: The current operation context.
        :type context: _ContextLike
        :raises NotImplementedError: This method must be implemented by subclasses.
        :return: The unpacked element.
        """
        raise NotImplementedError

    def pack_seq(self, seq, context) -> None:
        """
        Pack a sequence of elements using the provided context.

        :param seq: The sequence of elements to pack.
        :type seq: Iterable
        :param context: The current operation context.
        :type context: _ContextLike
        """
        pack_seq(seq, context, self.pack_single)

    def unpack_seq(self, context):
        """
        Unpack a sequence of elements using the provided context.

        :param context: The current operation context.
        :type context: _ContextLike
        :return: The list of unpacked elements.
        """
        return unpack_seq(context, self.unpack_single)

    def __pack__(self, obj, context) -> None:
        """
        Pack data based on whether the field is sequential or not.

        :param obj: The data to pack.
        :type obj: Any
        :param context: The current operation context.
        :type context: _ContextLike
        """
        (self.pack_single if not context[CTX_SEQ] else self.pack_seq)(obj, context)

    def __unpack__(self, context):
        """
        Unpack data based on whether the field is sequential or not.

        :param context: The current operation context.
        :type context: _ContextLike
        :return: The unpacked data.
        """
        return (self.unpack_seq if context[CTX_SEQ] else self.unpack_single)(context)

    def __repr__(self) -> str:
        """
        String representation of the FieldStruct instance.

        :return: A string representation.
        """
        return f"<{self.__class__.__name__}>"


class Chain(FieldStruct):
    """
    Represents a chain of structures where each structure in the chain is linked
    to the next one, forming a sequence.

    :param initial: The initial structure in the chain.
    :param structs: Additional structures to be added to the chain.

    The chain allows packing and unpacking data through its elements in sequence.

    .. note::
        - Unpacking travels from the head to the tail.
        - Packing travels from the tail to the head.
    """

    __slots__ = ("_elements",)

    def __init__(self, initial, *structs) -> None:
        # start -> next -> next -> next -> done | unpack
        #                                   Y
        # done <- previous <- previous <- start | pack
        self._elements = [getstruct(initial, initial)]
        self._elements += [x for x in map(lambda x: getstruct(x, x), structs) if x]

    @property
    def head(self):
        """
        Get the head of the chain, i.e., the first structure.

        :return: The head of the chain.
        :rtype: _StructLike
        """
        return self._elements[0]

    @property
    def tail(self):
        """
        Get the tail of the chain, i.e., the last structure.

        :return: The tail of the chain.
        :rtype: _StructLike
        """

        return self._elements[-1]

    def __size__(self, context) -> int:
        """
        Calculate the size of the chain in bytes.

        :param context: The context for the calculation.
        :type context: _ContextLike
        :return: The size of the chain.
        :rtype: int
        """
        return self.head.__size__(context)

    def __type__(self) -> type:
        """
        Get the type of the tail structure in the chain.

        :return: The type of the tail structure.
        :rtype: type
        """

        return self.tail.__type__()

    def __and__(self, other):
        """
        Concatenate another structure to the end of the chain.

        :param other: The structure to concatenate.
        :type other: _StructLike
        :return: The updated chain.
        :rtype: Chain
        """
        self._elements.append(getstruct(other, other))
        return self

    def __rand__(self, other):
        """
        Concatenate another structure to the beginning of the chain.

        :param other: The structure to concatenate.
        :type other: _StructLike
        :return: The updated chain.
        :rtype: Chain
        """
        return self.__and__(other)

    def unpack_single(self, context):
        """
        Unpack a single data instance from the chain.

        :param context: The context for the unpacking operation.
        :type context: _ContextLike
        :return: A memory view representing the unpacked data.
        :rtype: memoryview
        """
        data = None
        for i, struct in enumerate(self._elements):
            stream = BytesIO(data) if i != 0 else context[CTX_STREAM]
            with (
                WithoutContextVar(context, CTX_STREAM, stream),
                WithoutContextVar(context, CTX_SEQ, False),
            ):
                data = struct.__unpack__(context)

        return data

    def pack_single(self, obj, context) -> None:
        """
        Pack a single data instance into the chain.

        :param Any obj: The data to pack into the chain.
        :param context: The context for the packing operation.
        :type context: _ContextLike
        """
        count = len(self._elements)
        for i, struct in enumerate(reversed(self._elements)):
            if i == count - 1:
                # Last struct, use the provided context stream directly
                struct.__pack__(obj, context)
            else:
                # Not the last struct, use a temporary BytesIO object
                with (
                    BytesIO() as stream,
                    WithoutContextVar(context, CTX_STREAM, stream),
                    WithoutContextVar(context, CTX_SEQ, False),
                ):
                    struct.__pack__(obj, context)
                    obj = stream.getvalue()


class Operator:
    """Defines a custom opearator (user-defined)

    It operates _infix_ between two statements and takes them as
    agruments. For instance, the following example will return
    an array of structures:

    .. code-block:: python

        from caterpillar.fields import uint16, Operator
        from caterpillar.model import struct

        M = Operator(lambda a, b: a[b*2])

        @struct
        class Format:
            f1: uint16 /M/ 3

    This class reserves the `/` operator. It is also possible to
    use this class as a decorator on callable objects:

    .. code-block:: python

        @Operator
        def M(a, b):
            return a[b*2]

    :param func: The function to be applied.
    :type func: Callable[[Any, Any], _StructLike]
    """

    def __init__(self, func) -> None:
        self.func = func

    def __truediv__(self, arg2):
        return self.func(arg2)

    def __rtruediv__(self, arg1):
        return Operator(partial(self.func, arg1))

    def __call__(self, arg1, arg2):
        return self.func(arg1, arg2)


# utility methods
def get_args(args, context):
    """
    Get arguments for an instance.

    :param args: Input arguments.
    :type args: Any
    :param context: The current operation context.
    :type context: _ContextLike
    :return: A list of processed arguments.
    :rtype: list
    """
    args = list(args) if isinstance(args, Iterable) else [args]
    for i, argument in enumerate(args):
        if callable(argument):
            args[i] = argument(context)
    return args


def get_kwargs(kwargs: dict, context) -> dict:
    """
    Process a dictionary of keyword arguments, replacing callable values with their
    results.

    :param kwargs: Dictionary of keyword arguments.
    :type kwargs: dict
    :param context: The current operation context.
    :type context: _ContextLike
    :return: A new dictionary with processed keyword arguments.
    :rtype: dict
    """
    for key, value in kwargs.items():
        if callable(value):
            kwargs[key] = value(context)
    return kwargs



# Copyright (C) MatrixEditor 2023-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Any
from caterpillar.abc import _ContextLambda

# --- Shared Concepts ---
# TODO: This section needs some docs

# additional modifies set in the root context of each operation
MODE_PACK = 0
MODE_UNPACK = 1

# REVISIT: taken from reference
#: All models annotated with either @struct or @bitfield are struct
#: containers. Thus, they store the additional class attribute `__struct__`.
#:
#: Internally, any types utilizing this attribute can be employed within a
#: struct, bitfield, or sequence definition. The type of the stored value
#: must be conforming to the _StructLike protocol.
ATTR_STRUCT = "__struct__"
ATTR_BYTEORDER = "__byteorder__"
ATTR_TYPE = "__type__"
ATTR_BITS = "__bits__"
ATTR_SIGNED = "__signed__"
ATTR_TEMPLATE = "__template__"

# TODO: add to reference
# NEW CONCEPT: Actions
# An annotation that is equipped with an action attribute indicates that
# the struct will execute this very action instead of parsing the field.
#
# PROPOSAL:
#   For instance, an action could be used to install an IOHook on the
#   current stream (checksum or hash calculation).
#
#       class MyAction:
#           def __action__(self, context: _ContextLike) -> None:
#               ...
#
#       @struct
#       class MyStruct:
#           any_name: MyAction()
#
#   The action will be executed before the next field is parsed and won't
#   be stored in the model of the struct.
ATTR_ACTION_PACK = "__action_pack__"
ATTR_ACTION_UNPACK = "__action_unpack__"


class Action:
    """
    A class representing an action to be executed during the parsing or processing
    of a struct. This is used for cases where a field requires an operation to be
    performed instead of directly storing data.

    An action can be used to execute custom logic (such as modifying the IO stream)
    before the next field is processed. For example, it could be used to trigger
    checksum or hash calculations before reading a struct field.

    There are two types of actions:

    1. **Packing actions**: These actions are triggered before packing the data into the struct (i.e., before serializing or encoding).
    2. **Unpacking actions**: These actions are triggered before unpacking the data from the struct (i.e., before deserializing or decoding).

    Example:

    >>> def checksum_action(context: _ContextLike) -> None:
    ...     # This action could perform some checksum or logging logic
    ...     pass
    ...
    >>> @struct
    ... class MyStruct:
    ...     some_field: Bytes(10)
    ...     checksum: Action(checksum_action) # runs only when packing

    In this example, the `checksum` field is an action, and the `checksum_action`
    will be invoked before the `some_field` field is parsed.

    The action itself is not stored as part of the struct's model; it merely runs
    during struct processing.

    :param pack: The callable that will be executed before packing the struct (optional).
    :type pack: _ContextLambda | None
    :param unpack: The callable that will be executed before unpacking the struct (optional).
    :type unpack: _ContextLambda | None
    """

    __slots__ = (ATTR_ACTION_PACK, ATTR_ACTION_UNPACK)

    def __init__(self, pack=None, unpack=None, both=None) -> None:
        self.__action_pack__ = pack
        self.__action_unpack__ = unpack
        if both is not None:
            self.__action_pack__ = self.__action_unpack__ = both

    def __repr__(self) -> str:
        """
        Return a string representation of the action, showing its callable name
        and location.

        If only one of the action types is provided (either packing or unpacking),
        the corresponding action type (`PackAction` or `UnpackAction`) is displayed.

        :return: A string representation of the action.
        :rtype: str
        """
        name = "Action"
        if self.__action_pack__ and not self.__action_unpack__:
            name = "PackAction"
            return f"{name}({self.__action_pack__.__qualname__})"

        elif self.__action_unpack__ and not self.__action_pack__:
            name = "UnpackAction"
            return f"{name}({self.__action_unpack__.__qualname__})"

        if not self.__action_pack__ and not self.__action_unpack__:
            return name

        if self.__action_pack__ is self.__action_unpack__:
            return f"{name}({self.__action_pack__.__qualname__})"

        return f"{name}({self.__action_pack__.__qualname__}, {self.__action_unpack__.__qualname__})"

    @staticmethod
    def is_action(obj: Any) -> bool:
        """
        Checks if the given object is an instance of an `Action` class,
        based on the presence of the `__action_pack__` or `__action_unpack__` attributes.

        :param obj: The object to check.
        :type obj: Any
        :return: True if the object has either a `__action_pack__` or `__action_unpack__` attribute, indicating it is an action.
        :rtype: bool
        """
        return any(
            getattr(obj, attr, None) for attr in (ATTR_ACTION_PACK, ATTR_ACTION_UNPACK)
        )


def hasstruct(obj) -> bool:
    """
    Check if the given object has a structure attribute.

    :param obj: The object to check.
    :return: True if the object has a structure attribute, else False.
    """
    return hasattr(obj.__class__ if not isinstance(obj, type) else obj, ATTR_STRUCT)


def getstruct(obj, /, __default=None):
    """
    Get the structure attribute of the given object.

    :param obj: The object to get the structure attribute from.
    :return: The structure attribute of the object.
    """
    obj = obj.__class__ if not isinstance(obj, type) else obj
    return getattr(obj, ATTR_STRUCT, __default)


def typeof(struct):
    if hasstruct(struct):
        struct = getstruct(struct)

    __type__ = getattr(struct, ATTR_TYPE, None)
    if not __type__:
        return Any
    # this function must return a type
    return __type__() or Any
